"""
Graphs Functions
"""

def get_graphs(self, graphId=None, workspaceId=None, staged=False, cursor=None, limit=None, filters=None, fields=None):
    """Queries the workspace graphs based off provided parameters. If the workspaceId isn't specified, the current workspace will get used.
    
    Parameters
    ----------
    graphid : str
        GraphID to filter on. Optional.
    workspaceId : str    
        Workspace ID to filter on. If none is provided, the default workspace will get used.
    staged : bool
        If true, returns only graphs that are staged.
    cursor : str
        Cursor for pagination.
    limit : int
        Maximum number of graphs to return.
    filters: dict
        Filters that limit output to entries that match the filter 
    fields : list
        List of fields to return, leave empty to get all fields.
    
    Returns
    -------
    list[dict]
        A list of graphs based off provided query parameters if any parameters match.
    """
    self.check_logout()
    if workspaceId is None: workspaceId = self.workspace
    if limit is not None and limit <= 100: items = limit
    else: items = 100
    graphs = []
    while True:
        if limit and len(graphs) + items > limit: items = limit - len(graphs)
        ret = self.ana_api.getGraphs(workspaceId=workspaceId, graphId=graphId, staged=staged, limit=items, cursor=cursor, filters=filters, fields=fields)
        graphs.extend(ret)
        if len(ret) < items or len(graphs) == limit: break
        cursor = ret[-1]["graphId"]
    return graphs


def upload_graph(self, graph, channelId, name, description=None, staged=True, workspaceId=None):
    """Uploads a new graph based off provided parameters.
    
    Parameters
    ----------
    graph: str
        The graph as filepath, or python dictionary.
    channelId: str
        Id of channel to generate the graph with.
    name : str
        Name for the graph that will get generated.
    description: str
        Description of graph. Optional.
    staged: bool
        If true, the graph will get staged (read-only).
    workspaceId : str    
        Workspace ID create the graph in. If none is provided, the default workspace will get used. 
    
    Returns
    -------
    str
        The graphId if it was uploaded sucessfully.
    """
    import json
    import os
    import yaml

    self.check_logout()
    if graph is None: raise ValueError('The graph parameter is required!')
    if channelId is None: raise ValueError('The channelId parameter is required!')
    if name is None: raise ValueError('The name parameter is required!')
    if workspaceId is None: workspaceId = self.workspace
    if type(graph) is dict: graph = json.dumps(graph)
    else: 
        if not os.path.exists(graph): raise ValueError('The provided filepath for the graph parameter does not exist!')
        if graph.endswith('.json'): graph = open(graph, 'r').read()
        elif graph.endswith('.yaml') or graph.endswith('.yml'): graph = json.dumps(yaml.safe_load(open(graph, 'r').read()))
        else: raise ValueError('The graph parameter must be a dictionary or filepath of .json or .yaml file!')
    return self.ana_api.uploadGraph(workspaceId=workspaceId, channelId=channelId, graph=graph, name=name, description=description, staged=staged)


def edit_graph(self, graphId, name=None, description=None, graph=None, tags=None, workspaceId=None):
    """Update graph description and name. 
    
    Parameters
    ----------
    graphId : str
        Graph id to update.
    name: str
        New name to update.
    description: str
        New description to update.
    graph: str
        New graph to update.
    tags: list[str]
        New tags to update.
    workspaceId : str    
        Workspace ID of the graph's workspace. If none is provided, the current workspace will get used. 
    
    Returns
    -------
    bool
        If True, the graph was successfully edited.
    """
    self.check_logout()
    if graphId is None: raise ValueError('The graphId parameter is required!')
    if name is None and description is None and graph is None and tags is None: return True
    if workspaceId is None: workspaceId = self.workspace
    graphinfo = self.ana_api.getGraphs(workspaceId=workspaceId, graphId=graphId)[0]
    if graphinfo['staged'] and graph: raise ValueError('The graph is staged and cannot be edited!')
    return self.ana_api.editGraph(workspaceId=workspaceId, graphId=graphId, name=name, description=description, tags=tags)


def delete_graph(self, graphId, workspaceId=None):
    """Delete a graph in a workspace.
    
    Parameters
    ----------
    graphId : str
        Graph id to delete.
    workspaceId : str    
        Workspace ID of the graph's workspace. If none is provided, the current workspace will get used. 
    
    Returns
    -------
    str
        A success or error message based on graph's delete.
    """
    self.check_logout()
    if graphId is None: raise ValueError('The graphId parameter is required!')
    if workspaceId is None: workspaceId = self.workspace
    return self.ana_api.deleteGraph(workspaceId=workspaceId, graphId=graphId)
    

def download_graph(self, graphId, filepath=None, workspaceId=None):
    """Downloads a graph and save it to a file. If filepath is provided, the graph will get saved to that location.
    
    Parameters
    ----------
    graphId : str
        Graph ID of the graph to download.
    filepath : str
        Filepath to save the graph to. Optional.
    workspaceId : str    
        Workspace ID of the graph's workspace. If none is provided, the default workspace will get used. 
    
    Returns
    -------
    str
        The filepath of the downloaded graph.
    """
    import json
    import yaml

    self.check_logout()
    if graphId is None: raise ValueError('The graphId parameter is required!')
    if workspaceId is None: workspaceId = self.workspace
    graph = self.ana_api.getGraphs(workspaceId=workspaceId, graphId=graphId)[0]
    if filepath is None: filepath = f"{graph['name']}.yaml"
    graphstr = self.ana_api.downloadGraph(workspaceId=workspaceId, graphId=graphId)
    if filepath.endswith('.json'):
        with open(filepath, 'w+') as jsonfile:
            json.dump(json.loads(graphstr), jsonfile)
    elif filepath.endswith('.yaml') or filepath.endswith('.yml'):
        with open(filepath, 'w+') as yamlfile:
            yaml.safe_dump(json.loads(graphstr), yamlfile)
    else: raise ValueError('The filepath parameter must be a .json or .yaml file!')
    return filepath


def get_default_graph(self, channelId, filepath=None):
    """Downlaosd the default graph for a channel.
    
    Parameters
    ----------
    channelId:
        Id of channel to get the default graph for.
    filepath : str
        Filepath to save the graph to. Optional.

    Returns
    -------
    str
        The filepath of the downloaded graph.
    """
    import json
    import yaml
    
    self.check_logout()
    if channelId is None: raise ValueError('The channelId parameter is required!')
    graphstr = self.ana_api.getDefaultGraph(channelId=channelId)
    if filepath is None: filepath = "default.yaml"
    if filepath.endswith('.json'):
        with open(filepath, 'w+') as jsonfile:
            json.dump(json.loads(graphstr), jsonfile)
    elif filepath.endswith('.yaml') or filepath.endswith('.yml'):
        with open(filepath, 'w+') as yamlfile:
            yaml.safe_dump(json.loads(graphstr), yamlfile)
    else: raise ValueError('The filepath parameter must be a .json or .yaml file!')
    return filepath
    

def set_default_graph(self, graphId, workspaceId=None):
    """Sets the default graph for a channel. User must be in the organization that owns the channel.
    
    Parameters
    ----------
    graphId: str
        The ID of the graph that you want to be the default for the channel
    workspaceId : str
        The ID of the Workspace that the graph is in.

    Returns
    -------
    bool
        If True, the graph was successfully set as the default graph for the channel.
    """
    if self.check_logout(): return    
    if graphId is None: raise ValueError('The graphId parameter is required!')
    if workspaceId is None: workspaceId = self.workspace
    graph = self.ana_api.getGraphs(workspaceId=workspaceId, graphId=graphId)[0]
    return self.ana_api.setDefaultGraph(channelId=graph['channelId'], workspaceId=workspaceId, graphId=graphId)
    


"""
ML API calls.
"""

def getMLArchitectures(self, fields=None):
    if fields is None: fields = self.getTypeFields("MLArchitecture")
    fields = "\n".join(fields)
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "getMLArchitectures",
            "variables": {},
            "query": f"""query 
                getMLArchitectures {{
                    getMLArchitectures {{
                        {fields}
                    }}
                }}"""})
    return self.errorhandler(response, "getMLArchitectures")


def getMLModels(self, workspaceId, datasetId=None, modelId=None, cursor=None, limit=100, filters={}, fields=None):
    if fields is None: fields = self.getTypeFields("MLModel")
    fields = "\n".join(fields)
    if filters is None: filters = {}
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "getMLModels",
            "variables": {
                "workspaceId": workspaceId,
                "datasetId": datasetId,
                "modelId": modelId,
                "cursor": cursor,
                "limit": limit,
                "filters": filters
            },
            "query": f"""query 
                getMLModels ($workspaceId: String!, $datasetId: String, $modelId: String, $cursor: String, $limit: Int, $filters: MLModelFilter) {{
                    getMLModels (workspaceId: $workspaceId, datasetId: $datasetId, modelId: $modelId, cursor: $cursor, limit: $limit, filters: $filters) {{
                        {fields}
                    }}
                }}"""})
    return self.errorhandler(response, "getMLModels")


def createMLModel(self, workspaceId, datasetId, architectureId, name, description=None, parameters=None, tags=[]):
    response = self.session.post(
        url = self.url, 
        headers = self.headers,
        json = {
            "operationName": "createMLModel",
            "variables": {
                "workspaceId": workspaceId,
                "datasetId": datasetId,
                "architectureId": architectureId,
                "name": name,
                "description": description,
                "parameters": parameters,
                "tags": tags
            },
            "query": """mutation 
                createMLModel ($workspaceId: String!, $datasetId: String!, $architectureId: String!, $name: String, $description: String, $parameters: String, $tags: [String]) {
                    createMLModel (workspaceId: $workspaceId, datasetId: $datasetId, architectureId: $architectureId, name: $name, description: $description, parameters: $parameters, tags: $tags)
                }"""})
    return self.errorhandler(response, "createMLModel")


def deleteMLModel(self, workspaceId, modelId):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "deleteMLModel",
            "variables": {
                "workspaceId": workspaceId,
                "modelId": modelId,
            },
            "query": """mutation 
                deleteMLModel ($workspaceId: String!, $modelId: String!) {
                    deleteMLModel (workspaceId: $workspaceId, modelId: $modelId)
                }"""})
    return self.errorhandler(response, "deleteMLModel")


def editMLModel(self, workspaceId, modelId, name=None, description=None, tags=None):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "editMLModel",
            "variables": {
                "workspaceId": workspaceId,
                "modelId": modelId,
                "name": name,
                "description": description,
                "tags": tags
            },
            "query": """mutation 
                editMLModel ($workspaceId: String!, $modelId: String!, $name: String, $description: String, $tags: [String]) {
                    editMLModel (workspaceId: $workspaceId, modelId: $modelId, name: $name, description: $description, tags: $tags)
                }"""})
    return self.errorhandler(response, "editMLModel")


def downloadMLModel(self, workspaceId, modelId, checkpoint=None):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "downloadMLModel",
            "variables": {
                "workspaceId": workspaceId,
                "modelId": modelId,
                "checkpoint": checkpoint
            },
            "query": """mutation 
                downloadMLModel ($workspaceId: String!, $modelId: String!, $checkpoint: String) {
                    downloadMLModel (workspaceId: $workspaceId, modelId: $modelId, checkpoint: $checkpoint)
                }"""})
    return self.errorhandler(response, "downloadMLModel")


def uploadMLModel(self, workspaceId, architectureId, name, size, description=None, tags=[]):
    fields = self.getTypeFields("UploadMLModelResponse")
    fields = "\n".join(fields)
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "uploadMLModel",
            "variables": {
                "workspaceId": workspaceId,
                "architectureId": architectureId,
                "name": name,
                "size": size,
                "description": description,
                "tags": tags
            },
            "query": f"""mutation 
                uploadMLModel ($workspaceId: String!, $architectureId: String!, $name: String!, $size: Float!, $description: String, $tags: [String]) {{
                    uploadMLModel (workspaceId: $workspaceId, architectureId: $architectureId, name: $name, size: $size, description: $description, tags: $tags) {{
                        {fields}
                    }}
                }}"""})
    return self.errorhandler(response, "uploadMLModel")


def uploadMLModelFinalizer(self, workspaceId, key, uploadId, parts):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "uploadMLModelFinalizer",
            "variables": {
                "workspaceId": workspaceId,
                "key": key,
                "uploadId": uploadId,
                "parts": parts
            },
            "query": """mutation 
                uploadMLModelFinalizer ($workspaceId: String!, $key: String!, $uploadId: String, $parts: [MultipartInput!]) {
                    uploadMLModelFinalizer (workspaceId: $workspaceId, key: $key, uploadId: $uploadId, parts: $parts)
                }"""})
    return self.errorhandler(response, "uploadMLModelFinalizer")


def getMLInferences(self, workspaceId, inferenceId=None, datasetId=None, modelId=None, cursor=None, limit=100, filters={}, fields=None):
    if fields is None: fields = self.getTypeFields("MLInference")
    fields = "\n".join(fields)
    if filters is None: filters = {}
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "getMLInferences",
            "variables": {
                "workspaceId": workspaceId,
                "inferenceId": inferenceId,
                "datasetId": datasetId,
                "modelId": modelId,
                "cursor": cursor,
                "limit": limit,
                "filters": filters
            },
            "query": f"""query 
                getMLInferences ($workspaceId: String!, $inferenceId: String, $datasetId: String, $modelId: String, $cursor: String, $limit: Int, $filters: MLInferenceFilter) {{
                    getMLInferences (workspaceId: $workspaceId, inferenceId: $inferenceId, datasetId: $datasetId, modelId: $modelId, cursor: $cursor, limit: $limit, filters: $filters) {{
                        {fields}
                    }}
                }}"""})
    return self.errorhandler(response, "getMLInferences")


def getMLInferenceMetrics(self, workspaceId, inferenceId):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "getMLInferenceMetrics",
            "variables": {
                "workspaceId": workspaceId,
                "inferenceId": inferenceId
            },
            "query": """query 
                getMLInferenceMetrics ($workspaceId: String!, $inferenceId: String!) {
                    getMLInferenceMetrics (workspaceId: $workspaceId, inferenceId: $inferenceId)
                }"""})
    return self.errorhandler(response, "getMLInferenceMetrics")


def createMLInference(self, workspaceId, datasetId, modelId, mapId=None, tags=[]):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "createMLInference",
            "variables": {
                "workspaceId": workspaceId,
                "datasetId": datasetId,
                "modelId": modelId,
                "mapId": mapId,
                "tags": tags
            },
            "query": """mutation 
                createMLInference ($workspaceId: String!, $datasetId: String!, $modelId: String!, $mapId: String, $tags: [String]) {
                    createMLInference (workspaceId: $workspaceId, datasetId: $datasetId, modelId: $modelId, mapId: $mapId, tags: $tags)
                }"""})
    return self.errorhandler(response, "createMLInference")


def deleteMLInference(self, workspaceId, inferenceId):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "deleteMLInference",
            "variables": {
                "workspaceId": workspaceId,
                "inferenceId": inferenceId
            },
            "query": """mutation 
                deleteMLInference ($workspaceId: String!, $inferenceId: String!) {
                    deleteMLInference (workspaceId: $workspaceId, inferenceId: $inferenceId)
                }"""})
    return self.errorhandler(response, "deleteMLInference")


def editMLInference(self, workspaceId, inferenceId, tags=None):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "editMLInference",
            "variables": {
                "workspaceId": workspaceId,
                "inferenceId": inferenceId,
                "tags": tags
            },
            "query": """mutation 
                editMLInference ($workspaceId: String!, $inferenceId: String!, $tags: [String]) {
                    editMLInference (workspaceId: $workspaceId, inferenceId: $inferenceId, tags: $tags)
                }"""})
    return self.errorhandler(response, "editMLInference")


def downloadMLInference(self, workspaceId, inferenceId):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "downloadMLInference",
            "variables": {
                "workspaceId": workspaceId,
                "inferenceId": inferenceId
            },
            "query": """mutation 
                downloadMLInference ($workspaceId: String!, $inferenceId: String!) {
                    downloadMLInference (workspaceId: $workspaceId, inferenceId: $inferenceId)
                }"""})
    return self.errorhandler(response, "downloadMLInference")



"""
Volume Data Transfer Library

This module provides functions for efficient data transfer between volumes,
local paths, and S3 using direct boto3 operations with parallel execution.
"""

import os
import sys
import hashlib
import time
from datetime import datetime
from concurrent.futures import ThreadPoolExecutor, as_completed
from threading import Lock

# Check for boto3 dependency
try:
    import boto3
    from botocore.exceptions import ClientError
    BOTO3_AVAILABLE = True
except ImportError:
    BOTO3_AVAILABLE = False
    boto3 = None
    ClientError = Exception  # Placeholder

from anatools.lib.print import print_color


class TransferProgress:
    """Track and display transfer progress."""

    def __init__(self, total_files=0, total_size=0, verbose=False):
        self.total_files = total_files
        self.total_size = total_size
        self.completed_files = 0
        self.completed_size = 0
        self.failed_files = []
        self.start_time = time.time()
        self.lock = Lock()
        self.verbose = verbose

    def update(self, file_path, size, success=True):
        """Update progress for a completed file."""
        with self.lock:
            if success:
                self.completed_files += 1
                self.completed_size += size
            else:
                self.failed_files.append(file_path)
            self._display_progress()

    def _display_progress(self):
        """Display current progress."""
        if self.total_files == 0:
            return

        elapsed = time.time() - self.start_time
        percent = (self.completed_files / self.total_files) * 100

        # Calculate speed
        if elapsed > 0:
            speed_mb = (self.completed_size / (1024 * 1024)) / elapsed
            speed_str = f"{speed_mb:.2f} MB/s"
        else:
            speed_str = "calculating..."

        # Calculate ETA
        if self.completed_files > 0 and elapsed > 0:
            files_per_sec = self.completed_files / elapsed
            remaining_files = self.total_files - self.completed_files
            eta_seconds = remaining_files / files_per_sec
            eta_str = f"{int(eta_seconds)}s"
        else:
            eta_str = "calculating..."

        if not self.verbose:
            print(f"\rProgress: {self.completed_files}/{self.total_files} files ({percent:.1f}%) | "
                  f"Speed: {speed_str} | ETA: {eta_str}  ", end='', flush=True)
        else:
            print(f"Progress: {self.completed_files}/{self.total_files} files ({percent:.1f}%) | "
                  f"Speed: {speed_str} | ETA: {eta_str}")

    def finish(self):
        """Display final summary."""
        if not self.verbose:
            print()  # New line after progress bar
        elapsed = time.time() - self.start_time
        size_mb = self.completed_size / (1024 * 1024)

        print_color(f"\nTransfer completed!", 'brand')
        print(f"  Files transferred: {self.completed_files}/{self.total_files}")
        print(f"  Data transferred: {size_mb:.2f} MB")
        print(f"  Time elapsed: {elapsed:.2f}s")

        if self.failed_files:
            print_color(f"\n  Failed files ({len(self.failed_files)}):", 'error')
            for failed in self.failed_files[:10]:
                print(f"    - {failed}")
            if len(self.failed_files) > 10:
                print(f"    ... and {len(self.failed_files) - 10} more")


def get_volume_id_from_mountpoint(path):
    """
    Extract volume ID from a mount point by reading /proc/mounts.

    Args:
        path: Local path that might be a volume mount point

    Returns:
        str: Volume ID if found, None otherwise
    """
    import re

    abs_path = os.path.abspath(path)

    # Find the longest matching mount point
    try:
        with open('/proc/mounts', 'r') as f:
            mounts = f.readlines()

        best_match = None
        best_match_len = 0

        for line in mounts:
            parts = line.split()
            if len(parts) < 2:
                continue

            device = parts[0]
            mount_point = parts[1]

            # Check if our path is under this mount point
            if abs_path.startswith(mount_point):
                # Find the longest match (most specific mount point)
                if len(mount_point) > best_match_len:
                    best_match = (device, mount_point)
                    best_match_len = len(mount_point)

        if best_match:
            device, mount_point = best_match

            # Parse device path to extract volume ID
            # Format: bucket-name:/org-id/volume-id or bucket-name:/org-id/volume-id/
            # Example: renderedai-dev-volumes:/3da3f384-d6e8-45af-ab8f-bfee07948132/f084ad3b-2d32-4bf0-8bd5-52c0769b4811
            match = re.match(r'^[^:]+:/[a-f0-9\-]+/([a-f0-9\-]+)/?$', device)
            if match:
                volume_id = match.group(1)
                # Calculate prefix relative to mount point
                if abs_path == mount_point:
                    prefix = ''
                else:
                    prefix = os.path.relpath(abs_path, mount_point)
                return {
                    'volume_id': volume_id,
                    'prefix': prefix
                }

    except (IOError, OSError):
        pass

    return None


def parse_transfer_location(location):
    """
    Parse a transfer location string into its components.

    Args:
        location: String in format:
            - "volumeId" or "volumeId:prefix" for volumes
            - "/path/to/local" or "path/to/local" for local paths
            - Volume mount points are detected automatically

    Returns:
        dict: {
            'type': 'volume' or 'local',
            'volumeId': str (if volume) or None,
            'prefix': str (volume prefix or local path),
        }
    """
    if not location:
        raise ValueError("Location cannot be empty")

    # Check if it's a local path
    if location.startswith('/') or location.startswith('./') or location.startswith('..'):
        abs_path = os.path.abspath(location)

        # Check if this path is a volume mount point by reading /proc/mounts
        mount_info = get_volume_id_from_mountpoint(abs_path)
        if mount_info:
            return {
                'type': 'volume',
                'volumeId': mount_info['volume_id'],
                'prefix': mount_info['prefix']
            }

        # Regular local path
        return {
            'type': 'local',
            'volumeId': None,
            'prefix': abs_path
        }

    # Check for volume with prefix (contains colon)
    if ':' in location:
        parts = location.split(':', 1)
        volume_id = parts[0]
        prefix = parts[1].strip('/')  # Remove leading/trailing slashes

        # Validate volumeId format (basic check)
        if not volume_id or len(volume_id) < 5:
            # Might be a Windows path like C:\path, treat as local
            return {
                'type': 'local',
                'volumeId': None,
                'prefix': os.path.abspath(location)
            }

        return {
            'type': 'volume',
            'volumeId': volume_id,
            'prefix': prefix
        }

    # Check if it looks like a volume ID or a local path
    # Volume IDs are typically alphanumeric with dashes
    if '-' in location and not os.path.exists(location):
        return {
            'type': 'volume',
            'volumeId': location,
            'prefix': ''
        }

    # Default to local path
    return {
        'type': 'local',
        'volumeId': None,
        'prefix': os.path.abspath(location)
    }


def parse_s3_uri(s3_uri):
    """
    Parse an S3 URI into bucket and prefix.

    Args:
        s3_uri: S3 URI in format "s3://bucket/prefix/" or "bucket:/prefix/"

    Returns:
        dict: {
            'bucket': str,
            'prefix': str,
            'region': str
        }
    """
    # Parse S3 URI - handle both formats:
    # 1. s3://bucket-name/prefix/
    # 2. bucket-name:/prefix/
    if s3_uri.startswith('s3://'):
        s3_uri = s3_uri[5:]  # Remove 's3://'
        parts = s3_uri.split('/', 1)
        bucket = parts[0]
        prefix = parts[1].rstrip('/') if len(parts) > 1 else ''
    elif ':/' in s3_uri:
        # Handle bucket-name:/prefix/ format
        parts = s3_uri.split(':/', 1)
        bucket = parts[0]
        prefix = parts[1].rstrip('/') if len(parts) > 1 else ''
    else:
        raise Exception(f"Invalid S3 URI format: {s3_uri}")

    # Determine region (default to us-west-2 for Rendered.ai)
    region = 'us-west-2'

    return {
        'bucket': bucket,
        'prefix': prefix,
        'region': region
    }


def resolve_volume_s3_location(client, volume_id):
    """
    Get S3 bucket and prefix information for a volume using mount credentials.

    Args:
        client: anatools client instance
        volume_id: Volume ID to resolve

    Returns:
        dict: {
            'bucket': str,
            'prefix': str,
            'credentials': dict with 'accesskeyid', 'accesskey', 'sessiontoken',
            'region': str
        }
    """
    # Get mount credentials for the volume
    mount_data = client.mount_volumes(volumes=[volume_id])

    if not mount_data or 'credentials' not in mount_data:
        raise Exception(f"Failed to retrieve credentials for volume {volume_id}")

    # Parse S3 location from keys
    # keys format: ["s3://bucket-name/prefix/"]
    if 'keys' not in mount_data or not mount_data['keys']:
        raise Exception(f"No S3 keys found for volume {volume_id}")

    s3_uri = mount_data['keys'][0]
    s3_info = parse_s3_uri(s3_uri)

    return {
        'bucket': s3_info['bucket'],
        'prefix': s3_info['prefix'],
        'credentials': mount_data['credentials'],
        'region': s3_info['region']
    }


def create_s3_client(credentials, region='us-west-2'):
    """
    Create a boto3 S3 client with temporary credentials.

    Args:
        credentials: Dict with 'accesskeyid', 'accesskey', 'sessiontoken'
        region: AWS region

    Returns:
        boto3 S3 client
    """
    return boto3.client(
        's3',
        region_name=region,
        aws_access_key_id=credentials['accesskeyid'],
        aws_secret_access_key=credentials['accesskey'],
        aws_session_token=credentials['sessiontoken']
    )


def calculate_etag(file_path, chunk_size=8 * 1024 * 1024):
    """
    Calculate ETag for a file (MD5 hash).

    Args:
        file_path: Path to file
        chunk_size: Chunk size for reading

    Returns:
        str: MD5 hash (ETag)
    """
    md5 = hashlib.md5()
    with open(file_path, 'rb') as f:
        while True:
            chunk = f.read(chunk_size)
            if not chunk:
                break
            md5.update(chunk)
    return md5.hexdigest()


def list_s3_objects(s3_client, bucket, prefix, recursive=True):
    """
    List all objects in an S3 bucket with given prefix.

    Args:
        s3_client: boto3 S3 client
        bucket: S3 bucket name
        prefix: Key prefix
        recursive: If False, only list objects at current level

    Returns:
        list: List of dicts with 'key', 'size', 'etag', 'last_modified'
    """
    objects = []
    continuation_token = None

    # Add trailing slash to prefix if not empty
    if prefix and not prefix.endswith('/'):
        prefix += '/'

    while True:
        params = {
            'Bucket': bucket,
            'Prefix': prefix,
            'MaxKeys': 1000
        }

        if not recursive:
            params['Delimiter'] = '/'

        if continuation_token:
            params['ContinuationToken'] = continuation_token

        try:
            response = s3_client.list_objects_v2(**params)
        except ClientError as e:
            print_color(f"Error listing S3 objects: {e}", 'error')
            raise

        if 'Contents' in response:
            for obj in response['Contents']:
                # Skip directory markers
                if obj['Key'].endswith('/'):
                    continue

                objects.append({
                    'key': obj['Key'],
                    'size': obj['Size'],
                    'etag': obj['ETag'].strip('"'),
                    'last_modified': obj['LastModified']
                })

        if not response.get('IsTruncated'):
            break

        continuation_token = response.get('NextContinuationToken')

    return objects


def list_local_files(local_path, recursive=True):
    """
    List all files in a local directory.

    Args:
        local_path: Local directory path
        recursive: If False, only list files at current level

    Returns:
        list: List of dicts with 'path', 'size', 'relative_path'
    """
    files = []

    if not os.path.exists(local_path):
        raise Exception(f"Local path does not exist: {local_path}")

    if os.path.isfile(local_path):
        return [{
            'path': local_path,
            'size': os.path.getsize(local_path),
            'relative_path': os.path.basename(local_path)
        }]

    if recursive:
        for root, dirs, filenames in os.walk(local_path):
            for filename in filenames:
                file_path = os.path.join(root, filename)
                relative_path = os.path.relpath(file_path, local_path)
                files.append({
                    'path': file_path,
                    'size': os.path.getsize(file_path),
                    'relative_path': relative_path
                })
    else:
        for item in os.listdir(local_path):
            file_path = os.path.join(local_path, item)
            if os.path.isfile(file_path):
                files.append({
                    'path': file_path,
                    'size': os.path.getsize(file_path),
                    'relative_path': item
                })

    return files


def upload_file_to_s3(s3_client, local_file, bucket, key, max_retries=3):
    """
    Upload a single file to S3 with retries.

    Args:
        s3_client: boto3 S3 client
        local_file: Path to local file
        bucket: S3 bucket
        key: S3 key
        max_retries: Maximum number of retry attempts

    Returns:
        bool: True if successful, False otherwise
    """
    for attempt in range(max_retries):
        try:
            s3_client.upload_file(local_file, bucket, key)
            return True
        except ClientError as e:
            if attempt < max_retries - 1:
                time.sleep(2 ** attempt)  # Exponential backoff
            else:
                print_color(f"Failed to upload {local_file} to s3://{bucket}/{key}: {e}", 'error')
                return False
    return False


def download_file_from_s3(s3_client, bucket, key, local_file, max_retries=3):
    """
    Download a single file from S3 with retries.

    Args:
        s3_client: boto3 S3 client
        bucket: S3 bucket
        key: S3 key
        local_file: Path to save file locally
        max_retries: Maximum number of retry attempts

    Returns:
        bool: True if successful, False otherwise
    """
    # Create directory if it doesn't exist
    os.makedirs(os.path.dirname(local_file), exist_ok=True)

    for attempt in range(max_retries):
        try:
            s3_client.download_file(bucket, key, local_file)
            return True
        except ClientError as e:
            if attempt < max_retries - 1:
                time.sleep(2 ** attempt)  # Exponential backoff
            else:
                print_color(f"Failed to download s3://{bucket}/{key} to {local_file}: {e}", 'error')
                return False
    return False


def copy_s3_object(s3_client, source_bucket, source_key, dest_bucket, dest_key, max_retries=3):
    """
    Copy an object within S3 (server-side copy).

    Args:
        s3_client: boto3 S3 client
        source_bucket: Source S3 bucket
        source_key: Source S3 key
        dest_bucket: Destination S3 bucket
        dest_key: Destination S3 key
        max_retries: Maximum number of retry attempts

    Returns:
        bool: True if successful, False otherwise
    """
    copy_source = {'Bucket': source_bucket, 'Key': source_key}

    for attempt in range(max_retries):
        try:
            s3_client.copy_object(
                CopySource=copy_source,
                Bucket=dest_bucket,
                Key=dest_key
            )
            return True
        except ClientError as e:
            if attempt < max_retries - 1:
                time.sleep(2 ** attempt)  # Exponential backoff
            else:
                print_color(f"Failed to copy s3://{source_bucket}/{source_key} to s3://{dest_bucket}/{dest_key}: {e}", 'error')
                return False
    return False


def local_to_s3_transfer(client, source_path, dest_volume_id, dest_prefix='',
                         recursive=True, parallel=10, dry_run=False, verbose=False):
    """
    Transfer files from local path to S3 volume.

    Args:
        client: anatools client
        source_path: Local source path
        dest_volume_id: Destination volume ID
        dest_prefix: Destination prefix within volume
        recursive: Transfer recursively
        parallel: Number of parallel workers
        dry_run: Don't actually transfer, just show what would happen
        verbose: Verbose output

    Returns:
        bool: True if successful
    """
    print_color(f"Preparing upload from local to volume...", 'brand')

    # List local files
    local_files = list_local_files(source_path, recursive)

    if not local_files:
        print_color("No files found to transfer", 'warning')
        return True

    # Get S3 location for destination
    dest_s3 = resolve_volume_s3_location(client, dest_volume_id)
    dest_s3_client = create_s3_client(dest_s3['credentials'], dest_s3['region'])

    # Combine volume prefix and destination prefix
    if dest_prefix:
        full_dest_prefix = f"{dest_s3['prefix']}/{dest_prefix}".strip('/')
    else:
        full_dest_prefix = dest_s3['prefix']

    # Build transfer list
    transfer_list = []
    total_size = 0

    for file_info in local_files:
        dest_key = f"{full_dest_prefix}/{file_info['relative_path']}".strip('/')
        transfer_list.append({
            'local_file': file_info['path'],
            'dest_key': dest_key,
            'size': file_info['size']
        })
        total_size += file_info['size']

    if not transfer_list:
        print_color("No files need to be transferred", 'brand')
        return True

    print(f"Transferring {len(transfer_list)} files ({total_size / (1024*1024):.2f} MB)...")

    if dry_run:
        print_color("\n[DRY RUN] Would transfer:", 'warning')
        for item in transfer_list[:20]:
            print(f"  {item['local_file']} -> s3://{dest_s3['bucket']}/{item['dest_key']}")
        if len(transfer_list) > 20:
            print(f"  ... and {len(transfer_list) - 20} more files")
        return True

    # Execute parallel uploads
    progress = TransferProgress(len(transfer_list), total_size, verbose)

    with ThreadPoolExecutor(max_workers=parallel) as executor:
        futures = {}
        for item in transfer_list:
            future = executor.submit(
                upload_file_to_s3,
                dest_s3_client,
                item['local_file'],
                dest_s3['bucket'],
                item['dest_key']
            )
            futures[future] = item

        for future in as_completed(futures):
            item = futures[future]
            success = future.result()
            progress.update(item['local_file'], item['size'], success)

    progress.finish()
    return len(progress.failed_files) == 0


def s3_to_local_transfer(client, source_volume_id, source_prefix, dest_path,
                        recursive=True, parallel=10, dry_run=False, verbose=False):
    """
    Transfer files from S3 volume to local path.

    Args:
        client: anatools client
        source_volume_id: Source volume ID
        source_prefix: Source prefix within volume
        dest_path: Local destination path
        recursive: Transfer recursively
        parallel: Number of parallel workers
        dry_run: Don't actually transfer, just show what would happen
        verbose: Verbose output

    Returns:
        bool: True if successful
    """
    print_color(f"Preparing download from volume to local...", 'brand')

    # Get S3 location for source
    source_s3 = resolve_volume_s3_location(client, source_volume_id)
    source_s3_client = create_s3_client(source_s3['credentials'], source_s3['region'])

    # Combine volume prefix and source prefix
    if source_prefix:
        full_source_prefix = f"{source_s3['prefix']}/{source_prefix}".strip('/')
    else:
        full_source_prefix = source_s3['prefix']

    # List S3 objects
    s3_objects = list_s3_objects(source_s3_client, source_s3['bucket'], full_source_prefix, recursive)

    if not s3_objects:
        print_color("No files found to transfer", 'warning')
        return True

    # Build transfer list
    transfer_list = []
    total_size = 0

    for obj in s3_objects:
        # Calculate relative path by removing prefix
        rel_path = obj['key'][len(full_source_prefix):].lstrip('/')
        local_file = os.path.join(dest_path, rel_path)

        transfer_list.append({
            'source_key': obj['key'],
            'local_file': local_file,
            'size': obj['size']
        })
        total_size += obj['size']

    if not transfer_list:
        print_color("No files need to be transferred", 'brand')
        return True

    print(f"Transferring {len(transfer_list)} files ({total_size / (1024*1024):.2f} MB)...")

    if dry_run:
        print_color("\n[DRY RUN] Would transfer:", 'warning')
        for item in transfer_list[:20]:
            print(f"  s3://{source_s3['bucket']}/{item['source_key']} -> {item['local_file']}")
        if len(transfer_list) > 20:
            print(f"  ... and {len(transfer_list) - 20} more files")
        return True

    # Execute parallel downloads
    progress = TransferProgress(len(transfer_list), total_size, verbose)

    with ThreadPoolExecutor(max_workers=parallel) as executor:
        futures = {}
        for item in transfer_list:
            future = executor.submit(
                download_file_from_s3,
                source_s3_client,
                source_s3['bucket'],
                item['source_key'],
                item['local_file']
            )
            futures[future] = item

        for future in as_completed(futures):
            item = futures[future]
            success = future.result()
            progress.update(item['local_file'], item['size'], success)

    progress.finish()
    return len(progress.failed_files) == 0


def s3_to_s3_transfer(client, source_volume_id, source_prefix, dest_volume_id, dest_prefix,
                     recursive=True, parallel=10, dry_run=False, verbose=False):
    """
    Transfer files between S3 volumes (server-side copy).

    Args:
        client: anatools client
        source_volume_id: Source volume ID
        source_prefix: Source prefix within volume
        dest_volume_id: Destination volume ID
        dest_prefix: Destination prefix within volume
        recursive: Transfer recursively
        parallel: Number of parallel workers
        dry_run: Don't actually transfer, just show what would happen
        verbose: Verbose output

    Returns:
        bool: True if successful
    """
    print_color(f"Preparing volume-to-volume transfer (server-side copy)...", 'brand')

    # Get mount credentials for both volumes (single credential with access to both)
    mount_data = client.mount_volumes(volumes=[source_volume_id, dest_volume_id])

    if not mount_data or 'credentials' not in mount_data:
        raise Exception(f"Failed to retrieve credentials for volumes")

    # Parse S3 locations for both volumes from keys
    if 'keys' not in mount_data or len(mount_data['keys']) < 2:
        raise Exception(f"Expected 2 S3 keys, got {len(mount_data.get('keys', []))}")

    # Parse source S3 location
    source_s3_uri = mount_data['keys'][0]
    source_s3 = parse_s3_uri(source_s3_uri)

    # Parse destination S3 location
    dest_s3_uri = mount_data['keys'][1]
    dest_s3 = parse_s3_uri(dest_s3_uri)

    # Create S3 client with shared credentials
    s3_client = create_s3_client(mount_data['credentials'], source_s3['region'])

    # Combine volume prefixes with user-specified prefixes
    if source_prefix:
        full_source_prefix = f"{source_s3['prefix']}/{source_prefix}".strip('/')
    else:
        full_source_prefix = source_s3['prefix']

    if dest_prefix:
        full_dest_prefix = f"{dest_s3['prefix']}/{dest_prefix}".strip('/')
    else:
        full_dest_prefix = dest_s3['prefix']

    # List source objects
    source_objects = list_s3_objects(s3_client, source_s3['bucket'], full_source_prefix, recursive)

    if not source_objects:
        print_color("No files found to transfer", 'warning')
        return True

    # Build transfer list
    transfer_list = []
    total_size = 0

    for obj in source_objects:
        # Calculate relative path
        rel_path = obj['key'][len(full_source_prefix):].lstrip('/')
        dest_key = f"{full_dest_prefix}/{rel_path}".strip('/')

        transfer_list.append({
            'source_key': obj['key'],
            'dest_key': dest_key,
            'size': obj['size']
        })
        total_size += obj['size']

    if not transfer_list:
        print_color("No files need to be transferred", 'brand')
        return True

    print(f"Transferring {len(transfer_list)} files ({total_size / (1024*1024):.2f} MB)...")

    if dry_run:
        print_color("\n[DRY RUN] Would transfer:", 'warning')
        for item in transfer_list[:20]:
            print(f"  s3://{source_s3['bucket']}/{item['source_key']} -> s3://{dest_s3['bucket']}/{item['dest_key']}")
        if len(transfer_list) > 20:
            print(f"  ... and {len(transfer_list) - 20} more files")
        return True

    # Execute parallel S3 copies
    progress = TransferProgress(len(transfer_list), total_size, verbose)

    with ThreadPoolExecutor(max_workers=parallel) as executor:
        futures = {}
        for item in transfer_list:
            future = executor.submit(
                copy_s3_object,
                s3_client,
                source_s3['bucket'],
                item['source_key'],
                dest_s3['bucket'],
                item['dest_key']
            )
            futures[future] = item

        for future in as_completed(futures):
            item = futures[future]
            success = future.result()
            progress.update(item['dest_key'], item['size'], success)

    progress.finish()
    return len(progress.failed_files) == 0


def transfer_data(client, source, dest, recursive=True, parallel=10,
                 dry_run=False, verbose=False, **kwargs):
    """
    Main transfer orchestrator - routes to appropriate transfer method.

    Args:
        client: anatools client
        source: Source location (volumeId[:prefix] or path)
        dest: Destination location (volumeId[:prefix] or path)
        recursive: Transfer recursively
        parallel: Number of parallel workers
        dry_run: Don't actually transfer
        verbose: Verbose output
        **kwargs: Ignored (for backward compatibility)

    Returns:
        bool: True if successful
    """
    # Parse locations
    src = parse_transfer_location(source)
    dst = parse_transfer_location(dest)

    if verbose:
        print(f"Source: {src}")
        print(f"Destination: {dst}")

    # Route to appropriate transfer method
    if src['type'] == 'volume' and dst['type'] == 'volume':
        # Volume to volume (S3 to S3)
        return s3_to_s3_transfer(
            client, src['volumeId'], src['prefix'],
            dst['volumeId'], dst['prefix'],
            recursive, parallel, dry_run, verbose
        )
    elif src['type'] == 'local' and dst['type'] == 'volume':
        # Local to volume (upload)
        return local_to_s3_transfer(
            client, src['prefix'], dst['volumeId'], dst['prefix'],
            recursive, parallel, dry_run, verbose
        )
    elif src['type'] == 'volume' and dst['type'] == 'local':
        # Volume to local (download)
        return s3_to_local_transfer(
            client, src['volumeId'], src['prefix'], dst['prefix'],
            recursive, parallel, dry_run, verbose
        )
    else:
        # Local to local (not implemented - use cp/rsync)
        print_color("Local-to-local transfer not supported. Use 'cp' or 'rsync' instead.", 'error')
        return False



"""The client module is used for connecting to Rendered.ai's Platform API."""
import webbrowser
import http.server
import socketserver
import threading
import urllib
from urllib.parse import urlparse, parse_qs
import hashlib
import base64

envs = {
    'prod': {
        'name': 'Rendered.ai Platform',
        'url':  'https://deckard.rendered.ai',
        'statusAPI': 'https://api.rendered.ai/system',
        'api':  'https://api.rendered.ai/graphql',
        'auth': 'https://keycloak.rendered.ai' },
    'test': {
        'name': 'Rendered.ai Test Platform',
        'url':  'https://deckard.test.rendered.ai',
        'statusAPI': 'https://api.test.rendered.ai/system',
        'api':  'https://api.test.rendered.ai/graphql',
        'auth': 'https://keycloak.test.rendered.ai' },
    'dev': {
        'name': 'Rendered.ai Development Platform',
        'url':  'https://deckard.dev.rendered.ai/',
        'statusAPI': 'https://api.dev.rendered.ai/system',
        'api':  'https://api.dev.rendered.ai/graphql',
        'auth': 'https://keycloak.dev.rendered.ai' }
}

class AuthFailedError(Exception):
    """Custom exception for authentication failures."""
    pass

class client:

    def __init__(self, email=None, password=None, APIKey=None, bearer_token=None, environment=None, endpoint=None, local=False, interactive=True, verbose=None):
        from anatools.anaclient.api import api
        from anatools.lib.print import print_color
        from datetime import datetime
        import getpass
        import yaml
        import os
        import requests
        import anatools
        import time
        self.verbose = verbose
        self.interactive = interactive
        self.__bearer_token = bearer_token or os.environ.get('RENDEREDAI_BEARER_TOKEN')

        # check home directory for api key
        # Priority: Bearer Token -> API Key -> Email/Password
        if self.__bearer_token:
            pass # Handled after endpoint setup
        elif email is None and APIKey is None and os.environ.get('RENDEREDAI_API_KEY') is None:
            if os.path.exists(os.path.expanduser('~/.renderedai/config.yaml')):
                with open(os.path.expanduser('~/.renderedai/config.yaml'), 'r') as f:
                    config = yaml.safe_load(f)
                    if environment == None or environment == config['environment']:
                        APIKey = config['apikey']
                        environment = config['environment']
                        if self.interactive: print_color("Loaded API Key from ~/.renderedai/config.yaml", '91e600')

        # check environment
        if environment: self.environment = environment.lower()
        elif os.environ.get('RENDEREDAI_ENVIRONMENT'): self.environment = os.environ.get('RENDEREDAI_ENVIRONMENT').lower()
        else: self.environment = 'prod'
        if self.environment not in envs.keys():  raise Exception("Invalid environment argument.")

        # set client endpoints
        if local:
            os.environ['NO_PROXY'] = '127.0.0.1'
            self.__url = 'http://127.0.0.1:3000/graphql'
            self.__status_url = None
            self.__environment = 'Local'
            if self.interactive: print_color(f"Local is set to: {self.__url}", 'ffff00')
        elif endpoint:
            self.__url = endpoint
            self.__status_url = None
            self.__environment = 'Rendered.ai'
        elif os.environ.get('RENDEREDAI_ENDPOINT'):
            self.__url = os.environ.get('RENDEREDAI_ENDPOINT')
            self.__status_url = None
            self.__environment = 'Rendered.ai'
        else:
            self.__url = envs[self.environment]['api']
            self.__status_url = envs[self.environment]['statusAPI']
            self.__environment = envs[self.environment]['name']

        # Determine headers for API client
        initial_headers = None
        if self.__bearer_token:
            initial_headers = {'Authorization': f'Bearer {self.__bearer_token}'}

        self.ana_api = api(self.__url, self.__status_url, initial_headers, self.verbose)

        # initialize client variables
        self.__logout = False
        self.user = None
        self.organizations = None
        self.organization = None
        self.workspaces = None
        self.workspace = None
        self.channels = {}
        self.volumes = {}
        self.auth_method = None

        # configure client context
        if self.__bearer_token:
            try:
                # Fetch user context using the bearer token
                # This relies on getCurrentUserContext being available in self.ana_api (via api_keys.py)
                # and that it handles token errors via the errorhandler.
                user_context = self.ana_api.getCurrentUserContext()
                if not user_context or 'userId' not in user_context:
                    # errorhandler should have already printed a message if it's a token error
                    raise AuthFailedError("Failed to retrieve user context with Bearer Token.")

                self.user = user_context
                if 'userId' in self.user and 'uid' not in self.user:
                    self.user['uid'] = self.user.pop('userId')
                self.user['idtoken'] = self.__bearer_token # Store the token as idtoken for consistency
                # Ensure 'expiresAt' is handled if available from getCurrentUserContext, otherwise it might be None
                if 'expires' in self.user and 'expiresAt' not in self.user : # If 'expires' (duration) is returned
                    self.user['expiresAt'] = time.time() + self.user['expires']

                self.auth_method = 'bearer'
                self.__logout = False
                if self.interactive: print_color(f"Signed into {self.__environment} with Bearer Token.", '91e600')

                # Populate organizations and workspaces (similar to user-scope API key)
                self.organizations = self.get_organizations()
                organization = None
                if not self.organizations:
                    msg = "No organizations found for your account. Please contact support@rendered.ai."
                    if self.interactive: print_color(msg, 'ff0000')
                    raise AuthFailedError(msg)
                self.workspaces = self.get_workspaces()
                workspace = None
                if not self.workspaces:
                    response = input("No workspaces available. Would you like to create a new one? (y/n)")
                    if response.lower() == 'y':
                        self.create_workspace(name="Workspace")
                        self.workspaces = self.get_workspaces(fields=["workspaceId", "name", "organizationId"])
                        workspace = self.workspaces[0]
                        self.workspace = workspace['workspaceId']
                    else: raise Exception("No workspaces available. Please contact support@rendered.ai for support.")

                if os.getenv("RENDEREDAI_WORKSPACE_ID") and os.getenv("RENDEREDAI_WORKSPACE_ID") in [w['workspaceId'] for w in self.workspaces]: self.workspace = os.getenv("RENDEREDAI_WORKSPACE_ID")
                else: self.workspace = self.workspaces[0]['workspaceId']
                workspace = [w for w in self.workspaces if w['workspaceId'] == self.workspace][0] # Default to first workspace
                self.organization = workspace['organizationId']
                organization = [o for o in self.organizations if o['organizationId'] == workspace['organizationId']][0]

                if self.interactive and workspace:
                     print_color(f"The current organization is: {organization['name']}\nThe current workspace is: {workspace['name']}.", '91e600')
                elif self.interactive:
                     print_color(f"Set organization to {organization['name']}\nNo workspace set", 'ffff00')

            except AuthFailedError as e: # Catch specific auth errors from context fetching or setup
                if self.interactive: print_color(str(e), 'ff0000')
                raise
            except Exception as e: # Catch other unexpected errors during bearer token setup
                err_msg = f"An unexpected error occurred during Bearer Token authentication: {e}"
                if self.interactive: print_color(err_msg, 'ff0000')
                # The errorhandler in api.py should catch API call errors (like connection errors)
                # This handles errors in the client-side logic after a potentially successful API call
                raise AuthFailedError(err_msg)

        # If an email is provided, attempt email/password authentication.
        # Otherwise, the flow will fall through and trigger OAuth as the default.
        elif email:
            # The `email` argument is now required to enter this block, so we can remove the interactive prompt.
            self.__email = email
            # If a password is not provided and the session is interactive, prompt for it.
            if password is None and self.interactive:
                self.__password = getpass.getpass('Password: ')
            else:
                self.__password = password
            try:
                self.user = self.ana_api.login(email=self.__email, password=self.__password)
                if self.user: self.auth_method = 'password'
                if not self.user: raise AuthFailedError()
            except AuthFailedError as e:
                # If email/password login fails, print a message but don't raise an error.
                # This allows the client to fall back to the OAuth flow.
                if self.interactive: print_color(f'Failed to login to {self.__environment} with email {self.__email}.', 'ff0000')
                self.user = None
            except requests.exceptions.ConnectionError as e:
                print_color(f'Could not connect to API to login. Try again or contact support@rendered.ai for assistance.', 'ff0000')
                raise AuthFailedError()
            except requests.exceptions.JSONDecodeError as e:
                print_color(f'Failed to login with email {self.__email} and endpoint {self.__url}. Please confirm this is the correct email and endpoint, contact support@rendered.ai for assistance.', 'ff0000')
                raise AuthFailedError()
            self.ana_api = api(self.__url, self.__status_url, {'uid':self.user['uid'], 'Authorization': f'Bearer {self.user["idtoken"]}'}, self.verbose)
            try:
                response = self.ana_api.getSDKCompatibility()
                if response['version'] != anatools.__version__: print_color(response['message'], 'ff0000')
            except:
                if self.verbose: print_color("Failed to check SDK compatibility.", 'ff0000')
            # ask to create an api key if using email/password
            if self.interactive:
                resp = input('Would you like to create an API key to avoid logging in next time? (y/n): ')
                while resp.lower() not in ['y', 'n']:
                    resp = input('Invalid input, please respond with y or n: ')
                if resp.lower() == 'y':
                    print("What kind of scope would you like this API key to have?")
                    print("  [0] User - Full access to all organizations and workspaces you have access to.")
                    print("  [1] Organization - Access to a particular organization and any of it's workspaces.")
                    print("  [2] Workspace - Access to a particular workspace.")
                    resp = input('Please enter the number for the scope: ')
                    while resp.lower() not in ['0', '1', '2']:
                        resp = input('Invalid input, please respond with 0, 1, or 2: ')
                    datestr = datetime.now().isoformat()
                    if resp.lower() == '0':
                        apikey = self.create_api_key(name=f"anatools-{datestr}", scope='user')
                    elif resp.lower() == '1':
                        self.organizations = self.get_organizations()
                        organizations = [org for org in self.organizations if not org['expired']]
                        if len(organizations) == 0: raise Exception("No valid organizations found. Please contact sales@rendered.ai for support.")
                        print("Which organization would you like this API key to be associated with?")
                        for i, org in enumerate(organizations):
                            print(f"  [{i}] {org['name']}")
                        resp = input('Please enter a number for the organization: ')
                        while resp.lower() not in [str(i) for i in range(len(organizations))]:
                            resp = input(f"Invalid input, please respond with a number between 0 and {len(organizations)}: ")
                        self.organization = organizations[int(resp)]['organizationId']
                        apikey = self.create_api_key(name=f"anatools-{datestr}", scope='organization', organizationId=self.organization)
                    else:
                        self.organizations = self.get_organizations()
                        organizations = [org for org in self.organizations if not org['expired']]
                        if len(organizations) == 0: raise Exception("No valid organizations found. Please contact sales@rendered.ai for support.")
                        print("Which organization is the workspace is in?")
                        for i, org in enumerate(organizations):
                            print(f"  [{i}] {org['name']}")
                        resp = input('Please enter a number for the organization: ')
                        while resp.lower() not in [str(i) for i in range(len(organizations))]:
                            resp = input(f"Invalid input, please respond with a number between 0 and {len(organizations)}: ")
                        self.organization = organizations[int(resp)]['organizationId']
                        workspaces = self.get_workspaces(organizationId=self.organization, fields=["workspaceId", "name", "organizationId"])
                        if len(workspaces) == 0: raise Exception("No valid workspaces found in this organization. Please contact sales@rendered.ai for support.")
                        print("Which workspace would you like this API key to be associated with?")
                        for i, workspace in enumerate(workspaces):
                            print(f"  [{i}] {workspace['name']}")
                        resp = input('Please enter a number for the workspace: ')
                        while resp.lower() not in [str(i) for i in range(len(workspaces))]:
                            resp = input(f"Invalid input, please respond with a number between 0 and {len(workspaces)-1}: ")
                        self.workspace = workspaces[int(resp)]['workspaceId']
                        apikey = self.create_api_key(name=f"anatools-{datestr}", scope='workspace', workspaceId=self.workspace)
                    os.makedirs(os.path.expanduser('~/.renderedai'), exist_ok=True)
                    with open(os.path.expanduser('~/.renderedai/config.yaml'), 'w') as f:
                        yaml.dump({'apikey': apikey, 'environment': self.environment}, f)
                    print_color("API Key saved to ~/.renderedai/config.yaml", '91e600')

            if self.organization is None:
                self.organizations = self.get_organizations()
                if len(self.organizations) == 0: raise Exception("No organizations found. Please contact sales@rendered.ai for support.")
                organizations = [org for org in self.organizations if not org['expired']]
                if len(organizations) == 0: raise Exception("No valid organizations found. Please contact sales@rendered.ai for support.")
                self.organization = organizations[0]['organizationId']

            # get workspaces
            self.workspaces = self.get_workspaces(fields=["workspaceId", "name", "organizationId"])
            if self.workspace is None:
                if len(self.workspaces):
                    if os.getenv("RENDEREDAI_WORKSPACE_ID") and os.getenv("RENDEREDAI_WORKSPACE_ID") in [w['workspaceId'] for w in self.workspaces]: self.workspace = os.getenv("RENDEREDAI_WORKSPACE_ID")
                    else: self.workspace = self.workspaces[0]['workspaceId']
                    self.organization = [w for w in self.workspaces if w['workspaceId'] == self.workspace][0]['organizationId']
                else: raise Exception("No workspaces available. Please contact support@rendered.ai for support.")

            if self.interactive and self.workspace:
                workspace = [w for w in self.workspaces if w['workspaceId'] == self.workspace][0]['name']
                organization = [o for o in self.organizations if o['organizationId'] == self.organization][0]['name']
                print_color(f'Signed into {self.__environment} with {self.__email}.\nThe current organization is: {organization}\nThe current workspace is: {workspace}', '91e600')
        else:
            if APIKey or os.environ.get('RENDEREDAI_API_KEY'):
                if APIKey: self.__APIKey = APIKey
                else: self.__APIKey = os.environ.get('RENDEREDAI_API_KEY')
                if self.__APIKey:
                    self.sign_in_apikey()
                    self.auth_method = 'apikey'
                    if self.interactive:
                        print_color(f"Signed into {self.__environment} with API Key.", '91e600')
                        if self.organizations:
                            organization = [o for o in self.organizations if o['organizationId'] == self.organization][0]['name']
                            print_color(f'The current organization is: {organization}', '91e600')
                        if self.workspaces:
                            workspace = [w for w in self.workspaces if w['workspaceId'] == self.workspace][0]['name']
                            print_color(f'The current workspace is: {workspace}', '91e600')
            if not self.user:
                self._authenticate_oauth()


    def sign_in_apikey(self):
        from anatools.anaclient.api import api
        from anatools.lib.print import print_color
        from datetime import datetime
        import anatools
        import os
        import requests

        try:
            self.ana_api = api(self.__url, self.__status_url, {'apikey': self.__APIKey}, self.verbose)
            apikeydata = self.ana_api.getAPIKeyContext(apiKey=self.__APIKey)
            if not apikeydata:
                print_color("Invalid API Key", 'ff0000')
                raise AuthFailedError()
            if apikeydata.get('expiresAt'):
                apikey_date = datetime.strptime(apikeydata['expiresAt'], "%Y-%m-%dT%H:%M:%S.%fZ")
                current_date = datetime.now()
                if apikey_date < current_date:
                    print_color(f"API Key expired at {apikey_date}", 'ff0000')
                    raise AuthFailedError()
            try:
                response = self.ana_api.getSDKCompatibility()
                if response['version'] != anatools.__version__: print_color(response['message'], 'ff0000')
            except:
                if self.verbose: print_color("Failed to check SDK compatibility.", 'ff0000')
        except requests.exceptions.ConnectionError as e:
            raise Exception("Failed to reach Rendered.ai endpoint for login.")

        # workspace scope
        if apikeydata.get('workspaceId'):
            self.workspace = apikeydata['workspaceId']
            self.workspaces = self.get_workspaces(workspaceId=apikeydata['workspaceId'], fields=["workspaceId", "name", "organizationId"])
            self.organization = self.workspaces[0]['organizationId']
            self.user = {'status': 'authenticated_by_apikey'}
            return

        # organization scope
        elif apikeydata.get('organizationId'):
            self.organization = apikeydata['organizationId']
            self.organizations = self.get_organizations(organizationId=self.organization)
            self.workspaces = self.get_workspaces(organizationId=self.organization, fields=["workspaceId", "name", "organizationId"])
            if len(self.workspaces): 
                if os.getenv("RENDEREDAI_WORKSPACE_ID") and os.getenv("RENDEREDAI_WORKSPACE_ID") in [w['workspaceId'] for w in self.workspaces]: self.workspace = os.getenv("RENDEREDAI_WORKSPACE_ID")
                else: self.workspace = self.workspaces[0]['workspaceId']
            else:
                response = input("No workspaces available. Would you like to create a new one? (y/n)")
                if response.lower() == 'y':
                    self.create_workspace(name="Workspace", organizationId=self.organization)
                    self.workspaces = self.get_workspaces(organizationId=self.organization, fields=["workspaceId", "name", "organizationId"])
                    self.workspace = self.workspaces[0]['workspaceId']
                else: raise Exception("No workspaces available. Please contact support@rendered.ai for support.")
            self.user = {'status': 'authenticated_by_apikey'}

        # user scope
        else:
            self.organizations = self.get_organizations()
            if len(self.organizations): self.organization = self.organizations[0]['organizationId']
            else: raise Exception("No organizations found. Please contact sales@rendered.ai for support.")
            self.workspaces = self.get_workspaces(fields=["workspaceId", "name", "organizationId"])
            if len(self.workspaces):
                if os.getenv("RENDEREDAI_WORKSPACE_ID") and os.getenv("RENDEREDAI_WORKSPACE_ID") in [w['workspaceId'] for w in self.workspaces]: self.workspace = os.getenv("RENDEREDAI_WORKSPACE_ID")
                else: self.workspace = self.workspaces[0]['workspaceId']
            else:
                response = input("No workspaces available. Would you like to create a new one? (y/n)")
                if response.lower() == 'y':
                    self.create_workspace(name="Workspace")
                    self.workspaces = self.get_workspaces(fields=["workspaceId", "name", "organizationId"])
                    self.workspace = self.workspaces[0]['workspaceId']
                else: raise Exception("No workspaces available. Please contact support@rendered.ai for support.")
            self.organization = self.workspaces[0]['organizationId']
            self.user = {'status': 'authenticated_by_apikey'}

        validorgs = False
        for org in self.organizations:
            if not org['expired']: validorgs = True
        if not validorgs: raise Exception("No valid organizations found. Please contact sales@rendered.ai for support.")


    def refresh_token(self):
        import time
        from anatools.anaclient.api import api
        from anatools.lib.print import print_color

        # If auth_method is bearer or apikey, or if user is not set, do nothing
        if not hasattr(self, 'auth_method') or self.auth_method in ['bearer', 'apikey'] or not self.user:
            return

        # Existing refresh logic (which implies auth_method == 'password')
        if self.user:
            if int(time.time()) > int(self.user['expiresAt']):
                self.user = self.ana_api.login(self.__email, self.__password)
                self.ana_api = api(self.__url, self.__status_url, {'uid': self.user['uid'], 'Authorization': f'Bearer {self.user["idtoken"]}'}, self.verbose)
                try:
                    notification = self.ana_api.getSystemNotifications()
                    self.__notificationId = notification['notificationId']
                    if notification and notification['notificationId'] != self.__notificationId:
                        self.__notificationId = notification['notificationId']
                        print_color(notification['message'], 'ffff00')
                except requests.exceptions.ConnectionError as e:
                        print_color(f"Could not get notifications: {e}", 'ffff00')


    def check_logout(self):
        if self.__logout: raise Exception('You are currently logged out, login to access the Rendered.ai Platform.')
        self.refresh_token()


    def logout(self):
        """Logs out of the ana sdk and removes credentials from ana."""
        if self.check_logout(): return
        self.__logout = True
        del self.__password, self.__url, self.user


    def login(self, email=None, password=None, environment=None, endpoint=None, local=False, interactive=True, verbose=None):
        """Log in to the SDK.

        Parameters
        ----------
        email: str
            Email for the login. Will prompt if not provided.
        password: str
            Password to login. Will prompt if not provided.
        environment: str
            Environment to log into. Defaults to production.
        endpoint: str
            Custom endpoint to log into.
        local: bool
            Used for development to indicate pointing to local API.
        interactive: bool
            Set to False for muting the login messages.
        verbose: str
            Flag to turn on verbose logging. Use 'debug' to view log output.

        """
        self.__init__( email, password, environment, endpoint, local, interactive, verbose)


    def _authenticate_oauth(self):
        """
        Handles the OAuth2 authentication flow with PKCE.
        """
        from anatools.lib.print import print_color
        import os
        import requests
        import time

        client_id = 'anatools'
        realm = 'renderedai'
        keycloak_url = envs[self.environment]['auth']

        if not all([client_id, realm, keycloak_url]):
            if self.interactive:
                print_color("OAuth configuration is missing. Skipping OAuth flow.", 'ff0000')
            return

        # PKCE code verifier and challenge
        code_verifier = base64.urlsafe_b64encode(os.urandom(40)).decode('utf-8').rstrip('=')
        code_challenge = base64.urlsafe_b64encode(hashlib.sha256(code_verifier.encode('utf-8')).digest()).decode('utf-8').rstrip('=')

        redirect_uri = "http://localhost:9090"
        redirect_uri_enc = urllib.parse.quote(redirect_uri, safe='')
        auth_url = (f"{keycloak_url}/realms/{realm}/protocol/openid-connect/auth?"
                    f"client_id={client_id}&"
                    f"response_type=code&"
                    f"redirect_uri={redirect_uri_enc}&"
                    f"scope=openid%20email%20profile&"
                    f"code_challenge={code_challenge}&"
                    f"code_challenge_method=S256")

        print_color("No other authentication methods found, attempting to login via your browser.", 'ffff00')
        print_color(f"If your browser does not open automatically, please open this URL in your browser: \n{auth_url}", "91e600")
        webbrowser.open(auth_url)

        auth_code = [None]

        class OAuthCallbackHandler(http.server.SimpleHTTPRequestHandler):
            def log_message(self, format, *args):
                # Suppress default HTTP server logging
                return

            def do_GET(self):
                parsed_path = urlparse(self.path)
                query_params = parse_qs(parsed_path.query)
                if 'code' in query_params:
                    auth_code[0] = query_params['code'][0]
                    self.send_response(200)
                    self.send_header('Content-type', 'text/html')
                    self.end_headers()
                    self.wfile.write(b"<html><body><h1>Login successful!</h1><p>You can close this window.</p></body></html>")
                else:
                    self.send_response(400)
                    self.send_header('Content-type', 'text/html')
                    self.end_headers()
                    self.wfile.write(b"<html><body><h1>Login failed.</h1><p>No authorization code found.</p></body></html>")

                threading.Thread(target=self.server.shutdown).start()

        # Allow address reuse to prevent "Address already in use" errors
        socketserver.TCPServer.allow_reuse_address = True
        with socketserver.TCPServer(("", 9090), OAuthCallbackHandler) as httpd:
            httpd.serve_forever()

        if auth_code[0]:
            token_url = f"{keycloak_url}/realms/{realm}/protocol/openid-connect/token"
            token_data = {
                'grant_type': 'authorization_code',
                'client_id': client_id,
                'code': auth_code[0],
                'redirect_uri': redirect_uri,
                'code_verifier': code_verifier  # Add verifier for PKCE
            }

            try:
                if self.verbose == 'debug':
                    print_color(f"DEBUG: Token Request URL: {token_url}", '91e600')
                    print_color(f"DEBUG: Token Request Payload: {token_data}", '91e600')

                token_r = requests.post(token_url, data=token_data)

                if self.verbose == 'debug' and token_r.status_code != 200:
                    print_color(f"DEBUG: Token Error Response: {token_r.text}", 'ff0000')

                token_r.raise_for_status()
                token_json = token_r.json()
                self.__bearer_token = token_json['access_token']
                self.ana_api.headers['Authorization'] = f'Bearer {self.__bearer_token}'

                self.ana_api.session.headers.update(self.ana_api.headers)

                user_context = self.ana_api.getCurrentUserContext()
                if not user_context or 'userId' not in user_context:
                    raise AuthFailedError("Failed to retrieve user context with OAuth Token.")

                self.user = user_context
                if 'userId' in self.user and 'uid' not in self.user:
                    self.user['uid'] = self.user.pop('userId')
                self.user['idtoken'] = self.__bearer_token
                if 'expires_in' in token_json:
                    self.user['expiresAt'] = time.time() + token_json['expires_in']

                self.auth_method = 'oauth'
                self.__logout = False
                if self.interactive: print_color(f"Signed into {self.__environment} with OAuth.", '91e600')

                self.organizations = self.get_organizations()
                if not self.organizations:
                    msg = "No organizations found for your account. Please contact support@rendered.ai."
                    if self.interactive: print_color(msg, 'ff0000')
                    raise AuthFailedError(msg)
                self.workspaces = self.get_workspaces()

            except requests.exceptions.RequestException as e:
                if self.interactive:
                    print_color(f"Failed to exchange authorization code for a token: {e}", "ff0000")
                raise AuthFailedError("OAuth token exchange failed.")

    def get_system_status(self, serviceId=None, display=True):
        """Fetches the system status, if no serviceId is provided it will fetch all services.

        Parameters
        ----------
        serviceId: str
            The identifier of the service to fetch the status of.
        display: bool
            Boolean for either displaying the status or returning as a dict.
        """
        from anatools.lib.print import print_color
        services = self.ana_api.getSystemStatus(serviceId)
        if services and display:
            spacing = max([len(service['serviceName']) for service in services])+4
            print('Service Name'.ljust(spacing, ' ')+'Status')
            for service in services:
                print(service['serviceName'].ljust(spacing, ' '), end='')
                if service['status'] == 'Operational': print_color('Operational', '91e600')
                elif service['status'] == 'Degraded': print_color('Degraded', 'ffff00')
                elif service['status'] == 'Down': print_color('Down', 'ff0000')
                else: print('?')
            return
        return services




    from .organizations import set_organization, get_organizations, edit_organization, get_organization_members, get_organization_invites, add_organization_member, edit_organization_member, remove_organization_member, remove_organization_invitation
    from .workspaces    import set_workspace, get_workspaces, create_workspace, edit_workspace, delete_workspace, mount_workspaces, create_workspace_with_template, get_templates, create_template_request, get_template_requests
    from .graphs        import get_graphs, upload_graph, edit_graph, delete_graph, download_graph, get_default_graph, set_default_graph
    from .datasets      import get_datasets, get_dataset_jobs, create_dataset, edit_dataset, delete_dataset, download_dataset, cancel_dataset, upload_dataset, get_dataset_runs, get_dataset_log, get_dataset_files, create_mixed_dataset
    from .channels      import get_channels, get_channel_nodes, create_channel, edit_channel, delete_channel, build_channel, deploy_channel, get_deployment_status, get_channel_documentation, upload_channel_documentation, get_node_documentation
    from .volumes       import get_volumes, get_volumes, create_volume, edit_volume, delete_volume, get_volume_data, edit_volume_data, download_volume_data, upload_volume_data, delete_volume_data, mount_volumes, add_workspace_volumes, remove_workspace_volumes, search_volume
    from .analytics     import get_analytics, download_analytics, get_analytics_types, create_analytics, delete_analytics
    from .annotations   import get_annotations, get_annotation_formats, get_annotation_maps, create_annotation, download_annotation, delete_annotation , get_annotation_maps, upload_annotation_map, edit_annotation_map, delete_annotation_map, download_annotation_map
    from .gan           import get_gan_datasets, create_gan_dataset, delete_gan_dataset, get_gan_models, upload_gan_model, delete_gan_model, edit_gan_model, delete_gan_model, download_gan_model
    from .umap          import get_umaps, create_umap, delete_umap
    from .api_keys      import get_api_keys, create_api_key, delete_api_key
    from .llm           import get_llm_response, create_llm_prompt, delete_llm_prompt, get_llm_base_channels, get_llm_channel_node_types
    from .editor        import create_remote_development, delete_remote_development, list_remote_development, stop_remote_development, start_remote_development, prepare_ssh_remote_development, remove_ssh_remote_development, invite_remote_development, get_ssh_keys, register_ssh_key, deregister_ssh_key, get_servers, create_server, delete_server, edit_server, start_server, stop_server
    from .ml            import get_ml_architectures, get_ml_models, create_ml_model, delete_ml_model, edit_ml_model, download_ml_model, upload_ml_model, get_ml_inferences, get_ml_inference_metrics, create_ml_inference, delete_ml_inference, download_ml_inference
    from .inpaint       import get_inpaints, get_inpaint_log, create_inpaint, delete_inpaint
    from .preview       import get_preview, create_preview
    from .image         import get_image_annotation, get_image_mask, get_image_metadata
    from .agents        import get_data_types, get_data_fields
    from .services      import get_service_types, get_services, create_service, edit_service, delete_service, build_service, deploy_service, get_service_deployment, add_workspace_services, remove_workspace_services, get_service_jobs, create_service_job, delete_service_job, delete_service_job, get_workspace_service_credentials, get_instance_types
    from .rules         import get_platform_rules, get_organization_rules, get_workspace_rules, get_service_rules, get_user_rules, edit_organization_rules, edit_workspace_rules, edit_service_rules, edit_user_rules


"""
Agent Functions
"""

def get_data_types(self):
    """Retrieve a list of data types available on the Platform.
    
    Returns
    -------
    list[str]
        The data types available on the Platform.
    """
    return self.ana_api.getTypes()


def get_data_fields(self, type):
    """Retrieve information about the fields that can be returned for a specific data type.
    
    Parameters
    ----------
    type : str
        The data type to retrieve fields for.
    
    Returns
    -------
    list[str]
        Information about the fields available for the specified data type.
    """
    if type not in self.get_data_types(): raise ValueError(f"Invalid data type: {type}")
    return self.ana_api.getTypeFields(type)



"""
LLM API calls.
"""

def getLLMResponse(self, promptId, fields=None):
    if fields is None: fields = self.getTypeFields("LLMResponse")
    fields = "\n".join(fields)
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "getLLMResponse",
            "variables": {
                "promptId": promptId
            },
            "query": f"""query 
                getLLMResponse($promptId: String!) {{
                    getLLMResponse(promptId: $promptId) {{
                        {fields}
                    }}
                }}"""})
    return self.errorhandler(response, "getLLMResponse")


def createLLMPrompt(self, prompt, baseChannel, nodeType, nodeName):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "createLLMPrompt",
            "variables": {
                "prompt": prompt,
                "baseChannel": baseChannel,
                "nodeType": nodeType,
                "nodeName": nodeName
            },
            "query": """mutation 
                createLLMPrompt($prompt: String!, $baseChannel: String!, $nodeType: String!, $nodeName: String!) {
                    createLLMPrompt(prompt: $prompt, baseChannel: $baseChannel, nodeType: $nodeType, nodeName: $nodeName)
                }"""})
    return self.errorhandler(response, "createLLMPrompt")


def deleteLLMPrompt(self, promptId):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "createLLMPrompt",
            "variables": {
                "promptId": promptId
            },
            "query": """mutation 
                deleteLLMPrompt($promptId: String!) {
                    deleteLLMPrompt(promptId: $promptId)
                }"""})
    return self.errorhandler(response, "deleteLLMPrompt")


def getLLMBaseChannels(self):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "getLLMBaseChannels",
            "variables": {},
            "query": """query 
                getLLMBaseChannels {
                    getLLMBaseChannels
                }"""})
    return self.errorhandler(response, "getLLMBaseChannels")


def getLLMChannelNodeTypes(self):
    fields = self.getTypeFields("LLMChannelNodeTypes")
    fields = "\n".join(fields)
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "getLLMChannelNodeTypes",
            "variables": {},
            "query": f"""query 
                getLLMChannelNodeTypes {{
                    getLLMChannelNodeTypes {{
                        {fields}
                    }}
                }}"""})
    return self.errorhandler(response, "getLLMChannelNodeTypes")



"""
Channels API calls.
"""

def getChannels(self, organizationId=None, workspaceId=None, channelId=None, limit=100, cursor=None, filters={}, fields=None):
    if fields is None: fields = self.getTypeFields("Channel")
    fields = "\n".join(fields)
    if filters is None: filters = {}
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "getChannels",
            "variables": {
                "organizationId": organizationId,
                "workspaceId": workspaceId,
                "channelId": channelId,
                "limit": limit,
                "cursor": cursor,
                "filters": filters
            },
            "query": F"""query 
                getChannels($organizationId: String, $workspaceId: String, $channelId: String, $limit: Int, $cursor: String, $filters: ChannelFilter) {{
                    getChannels(organizationId: $organizationId, workspaceId: $workspaceId, channelId: $channelId, limit: $limit, cursor: $cursor, filters: $filters) {{
                        {fields}
                    }}
                }}"""})
    return self.errorhandler(response, "getChannels")


def getChannelDeployment(self, deploymentId):
    fields = self.getTypeFields("ChannelDeployment")
    fields = "\n".join(fields)
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "getChannelDeployment",
            "variables": {
                "deploymentId": deploymentId
            },
            "query": f"""query 
                getChannelDeployment($deploymentId: String!) {{
                    getChannelDeployment(deploymentId: $deploymentId) {{
                        {fields}
                    }}
                }}"""})
    return self.errorhandler(response, "getChannelDeployment")


def getChannelSchema(self, channelId):
    response = self.session.post(
        url = self.url, 
        headers = self.headers,
        json = {
            "operationName": "getChannelSchema",
            "variables": {
                "channelId": channelId
            },
            "query": """query getChannelSchema($channelId: String) {
                getChannelSchema(channelId: $channelId)
            }"""})
    return self.errorhandler(response, "getChannelSchema")


def createChannel(self, organizationId, name, description=None, volumes=None, instance=None, timeout=None, interfaceVersion=None):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "createChannel",
            "variables": {
                "organizationId": organizationId,
                "name": name,
                "description": description,
                "volumes": volumes,
                "instance": instance,
                "timeout": timeout,
                "interfaceVersion": interfaceVersion
            },
            "query": """mutation 
                createChannel($organizationId: String!, $name: String!, $description: String, $volumes: [String], $instance: String, $timeout: Int, $interfaceVersion: Int) {
                    createChannel(organizationId: $organizationId, name: $name, description: $description, volumes: $volumes, instance: $instance, timeout: $timeout, interfaceVersion: $interfaceVersion)
                }"""})
    return self.errorhandler(response, "createChannel")


def deleteChannel(self, channelId, organizationId=None):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "deleteChannel",
            "variables": {
                "channelId": channelId,
                "organizationId": organizationId
            },
            "query": """mutation 
                deleteChannel($channelId: String!, $organizationId: String!) {
                    deleteChannel(channelId: $channelId, organizationId: $organizationId)
                }"""})
    return self.errorhandler(response, "deleteChannel")


def editChannel(self, channelId, name=None, description=None, volumes=None, instance=None, timeout=None, status=None, interfaceVersion=None, preview=None):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "editChannel",
            "variables": {
                "channelId": channelId,
                "name": name,
                "description": description,
                "volumes": volumes,
                "instance": instance,
                "timeout": timeout,
                "status":status,
                "interfaceVersion": interfaceVersion,
                "preview": preview
            },
            "query": """mutation 
                editChannel($channelId: String!, $name: String, $description: String, $volumes: [String], $instance: String, $timeout: Int, $status: String, $interfaceVersion: Int, $preview: Boolean) {
                    editChannel(channelId: $channelId, name: $name, description: $description, volumes: $volumes, instance: $instance, timeout: $timeout, status: $status, interfaceVersion: $interfaceVersion, preview: $preview)
                }"""})
    return self.errorhandler(response, "editChannel")


def deployChannel(self, channelId, alias=None):
    fields = self.getTypeFields("ECRDeployment")
    fields = "\n".join(fields)
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "deployChannel",
            "variables": {
                "channelId": channelId,
                "alias": alias
            },
            "query": f"""mutation 
                deployChannel($channelId: String!, $alias: String) {{
                    deployChannel(channelId: $channelId, alias: $alias) {{
                        {fields}
                    }}
                }}"""})
    return self.errorhandler(response, "deployChannel")

    
def getChannelDocumentation(self, channelId):
    response = self.session.post(
        url = self.url, 
        headers = self.headers,
        json = {
            "operationName": "getChannelDocumentation",
            "variables": {
                "channelId": channelId
            },
            "query": """query getChannelDocumentation($channelId: String!) {
                getChannelDocumentation(channelId: $channelId)
            }"""})
    return self.errorhandler(response, "getChannelDocumentation")


def uploadChannelDocumentation(self, channelId, keys=[]):
    fields = self.getTypeFields("ChannelDocumentationUpload")
    fields = "\n".join(fields)
    response = self.session.post(
        url = self.url, 
        headers = self.headers,
        json = {
            "operationName": "uploadChannelDocumentation",
            "variables": {
                "channelId": channelId,
                "keys": keys,
            },
            "query": f"""mutation uploadChannelDocumentation($channelId: String!, $keys: [String]!) {{
                uploadChannelDocumentation(channelId: $channelId, keys: $keys) {{
                    {fields}
                }}
            }}"""})
    return self.errorhandler(response, "uploadChannelDocumentation")


def getNodeDocumentation(self, channelId, node, fields=None):
    if fields is None: fields = self.getTypeFields("NodeDocumentation")
    fields = "\n".join(fields)
    response = self.session.post(
        url = self.url, 
        headers = self.headers,
        json = {
            "operationName": "getNodeDocumentation",
            "variables": {
                "channelId": channelId,
                "nodeClass": node
            },
            "query": f"""query getNodeDocumentation($channelId: String!, $nodeClass: String!) {{
                getNodeDocumentation(channelId: $channelId, nodeClass: $nodeClass) {{
                    {fields}
                }}
            }}"""})
    return self.errorhandler(response, "getNodeDocumentation")


"""API Module"""

class api:

    def __init__(self, url, status_url, headers, verbose=False):
        import requests
        self.url = url
        self.status_url = status_url
        self.headers = headers or {}
        self.verbose = verbose
        self.types = None
        self.fields = {}
        self.session = requests.Session()
        if self.headers:
            self.session.headers.update(self.headers)

    def login(self, email, password):
        import time
        fields = self.getTypeFields("UserCredentials")
        fields = "\n".join(fields)
        response = self.session.post(
            url = self.url,
            json = {
                "operationName": "signIn",
                "variables": {
                    "email": email,
                    "password": password
                },
                "query": f"""mutation
                    signIn($email: String!, $password: String!) {{
                        signIn(email: $email, password: $password) {{
                            {fields}
                        }}
                    }}"""})
        if 'errors' in response.json(): return False
        data = self.errorhandler(response, "signIn")
        data['expiresAt'] = time.time() + data['expires']
        return data


    def close(self):
        self.session.close()


    def getSystemNotifications(self):
        if self.status_url is None: return None
        response = self.session.post(
            url = self.status_url,
            json = {
                "operationName": "getSystemNotifications",
                "variables": {},
                "query": """query
                    getSystemNotifications {
                        getSystemNotifications {
                            message
                            notificationId
                        }
                    }"""})
        if 'errors' in response.json(): return False
        return self.errorhandler(response, "getSystemNotifications")


    def getSystemStatus(self, serviceId=None):
        if self.status_url is None: return None
        response = self.session.post(
            url = self.status_url,
            json = {
                "operationName": "getSystemStatus",
                "variables": {
                    "serviceId": serviceId
                },
                "query": """query
                    getSystemStatus($serviceId: String) {
                        getSystemStatus(serviceId: $serviceId) {
                            serviceId
                            serviceName
                            description
                            status
                            type
                            updatedAt
                            createdAt
                        }
                    }"""})
        if 'errors' in response.json(): return False
        return self.errorhandler(response, "getSystemStatus")


    def getSDKCompatibility(self):
        import anatools
        import platform
        os = str(platform.system_alias(platform.system(), platform.release(), platform.version()))
        python = str(platform.python_version())
        anatools = str(anatools.__version__)
        response = self.session.post(
            url = self.url,
            json = {
                "operationName": "getSDKCompatibility",
                "variables": {
                    "os": os,
                    "python": python,
                    "anatools": anatools
                },
                "query": """query
                    getSDKCompatibility($os: String!, $python: String!, $anatools: String!) {
                        getSDKCompatibility(os: $os, python: $python, anatools: $anatools) {
                            version
                            message
                        }
                    }"""
            })
        if 'errors' in response.json(): return False
        return self.errorhandler(response, "getSDKCompatibility")


    def errorhandler(self, response, call):
        responsedata = response.json()
        if self.verbose == 'debug': print(responsedata)
        try:
            if 'data' in responsedata and responsedata['data'] is not None and call in responsedata['data']: return responsedata['data'][call]
            elif 'errors' in responsedata: raise Exception(responsedata['errors'][-1]['message'])
            else: raise Exception()
        except Exception as e:
            raise Exception(f'There was an issue with the {call} API call: {e}')


    from .organizations import getOrganizations, editOrganization
    from .channels      import getChannels, getChannelDeployment, getChannelSchema, createChannel, deleteChannel, editChannel, deployChannel, getChannelDocumentation, uploadChannelDocumentation, getNodeDocumentation
    from .volumes       import getVolumes, getVolumes, createVolume, deleteVolume, editVolume, editVolumeData, getVolumeData, uploadVolumeData, uploadVolumeDataFinalizer, deleteVolumeData, mountVolumes, addWorkspaceVolumes, removeWorkspaceVolumes
    from .members       import getMembers, addMember, removeMember, editMember, getInvitations
    from .workspaces    import getWorkspaces, createWorkspace, deleteWorkspace, editWorkspace, mountWorkspaces, createWorkspaceWithTemplate, getTemplates, createTemplateRequest, getTemplateRequests   
    from .graphs        import getGraphs, uploadGraph, deleteGraph, editGraph, downloadGraph, getDefaultGraph, setDefaultGraph
    from .datasets      import getDatasets, getDatasetJobs, createDataset, deleteDataset, editDataset, downloadDataset, cancelDataset, uploadDataset, uploadDatasetFinalizer, getDatasetRuns, getDatasetLog, getDatasetFiles, createMixedDataset
    from .analytics     import getAnalytics, getAnalyticsTypes, createAnalytics, deleteAnalytics, editAnalytics
    from .annotations   import getAnnotations, getAnnotationFormats, getAnnotationMaps, createAnnotation, downloadAnnotation, deleteAnnotation, editAnnotation, getAnnotationMaps, uploadAnnotationMap, uploadAnnotationMapFinalizer, editAnnotationMap, deleteAnnotationMap, downloadAnnotationMap
    from .gan           import getGANModels, getGANDatasets, createGANDataset, deleteGANDataset, uploadGANModel, uploadGANModelFinalizer, deleteGANModel, editGANModel, downloadGANModel
    from .umap          import getUMAPs, createUMAP, deleteUMAP
    from .api_keys      import getAPIKeys, createAPIKey, deleteAPIKey, getAPIKeyContext, getCurrentUserContext
    from .llm           import getLLMResponse, createLLMPrompt, deleteLLMPrompt, getLLMBaseChannels, getLLMChannelNodeTypes
    from .editor        import createRemoteDevelopment, deleteRemoteDevelopment, listRemoteDevelopment, startRemoteDevelopment, stopRemoteDevelopment, inviteRemoteDevelopment, createSSHKey, deleteSSHKey, getSSHKeys, getEditors, createEditor, deleteEditor, editEditor, startEditor, stopEditor
    from .ml            import getMLArchitectures, getMLModels, createMLModel, deleteMLModel, editMLModel, downloadMLModel, uploadMLModel, uploadMLModelFinalizer, getMLInferences, getMLInferenceMetrics, createMLInference, deleteMLInference, editMLInference, downloadMLInference
    from .inpaint       import getInpaints, getInpaintLog, createInpaint, deleteInpaint
    from .preview       import getPreview, createPreview
    from .image         import getImageAnnotation, getImageMask, getImageMetadata
    from .introspection import getTypes, getTypeFields
    from .services      import getServiceTypes, getServices, createService, editService, deleteService, deployService, getServiceDeployment, addWorkspaceServices, removeWorkspaceServices, getServiceJobs, createServiceJob, deleteServiceJob, getWorkspaceServiceCredentials, getInstanceTypes
    from .rules         import getPlatformRules,getOrganizationRules, getWorkspaceRules, getServiceRules, getUserRules, editOrganizationRules, editWorkspaceRules, editServiceRules, editUserRules


# Copyright 2019-2022 DADoES, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License in the root directory in the "LICENSE" file or at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import yaml
import glob
import logging
import os
import sys
import importlib
from pathlib import Path
import anatools
import anatools.lib.context as ctx

logger = logging.getLogger(__name__)


def find_servicefile():
    service = None
    servicefiles = []
    # search local directory
    localfiles = [file for file in os.listdir('.') if file.endswith('.yml') or file.endswith('.yaml') ]
    for file in localfiles: servicefiles.append(file)
    for servicefile in servicefiles:
        with open(servicefile, 'r') as f:
            cfg = yaml.safe_load(f)
            # if it adds packages then assume it's a service file
            if "service" in cfg:
                service = servicefile
                print(f'Using servicefile found at {service}.\nIf this is the wrong service, specify a servicefile using the --service argument.')
                break
    return service


class Service:
    
    def __init__(self, servicefile):
        """ Create a service class from a service file """
        self.servicefile = servicefile
        self.description = None
        self.volumes = []
        self.schemas = {}
        self.remotes = []
        with open(servicefile, 'r') as f:
            cfg = yaml.safe_load(f)
            if "service" in cfg:
                if 'name' in cfg["service"]: self.name = cfg["service"].get("name", os.path.dirname(servicefile))
                else: self.name = os.path.dirname(servicefile)
                if 'description' in cfg["service"]: self.description = cfg["service"].get("description", None)
                else: self.description = None
                if 'volumes' in cfg["service"]: self.volumes = cfg["service"].get("volumes", [])
                else: self.volumes = []
                if 'tools' in cfg["service"]: self.schemas = {"tools": cfg["service"].get("tools", {})}
                elif 'execs' in cfg["service"]: self.schemas = {"execs": cfg["service"].get("execs", {})}
                else: self.schemas = {"tools": {}}
            if "remotes" in cfg:
                self.remotes = cfg["remotes"]
                



"""
Services Functions
"""

def get_service_types(self, fields=None):
    """Get the service types supported by the platform.
    
    Parameters
    ----------
    fields : list
        List of fields to return, leave empty to get all fields.
   
    Returns
    -------
    list[dict]
        service types
    """
    import json
    self.check_logout()
    response = self.ana_api.getServiceTypes(fields=fields)
    return response


def get_services(self, organizationId=None, workspaceId=None, serviceId=None, cursor=None, limit=None, filters=None, fields=None):
    """Fetches all services available to an organization or workspace.
    
    Parameters
    ----------
    organizationId : str
        Filter services list on what's available to the organization.
    workspaceId : str    
        Filter services list on what's available to the workspace.
    serviceId: str
        Filter services list on the specific serviceId.
    cursor : str
        Cursor for pagination.
    limit : int
        Maximum of services to return.
    filters: dict
        Filters that limit output to entries that match the filter
    fields : list
        List of fields to return, leave empty to get all fields.
    
    Returns
    -------
    list[dict]
        List of services associated with workspace, organization or serviceId.
    """
    self.check_logout()
    if limit is not None and limit <= 100: items = limit
    else: items = 100
    services = []
    while True:
        if limit and len(services) + items > limit: items = limit - len(services)
        ret = self.ana_api.getServices(organizationId=organizationId, workspaceId=workspaceId, serviceId=serviceId, limit=items, cursor=cursor, filters=filters, fields=fields)
        services.extend(ret)
        if len(ret) < items or len(services) == limit: break
        cursor = ret[-1]["serviceId"]
    return services


def create_service(self, name, description=None, organizationId=None, serviceTypeId=None, volumes=[], instance=None, tags=[], serverId=None):
    """Create a new service for your organization.
    
    Parameters
    ----------
    name : str
        Service name.
    description : str
        Description of the service
    organizationId : str
        Organization ID. Defaults to current if not specified.
    serviceType : str
        The service type that the service adheres to.
    volumes : list[str]
        List of the data volume names to associate with this service.
    instance: str
        AWS Instance type the service will run on.
    tags : list[str]
        List of tags to associate with this service.
    serverId: str
        The ID of the server the service is drafted on, if applicable.
   
    Returns
    -------
    str
        ServiceId of the new service.
    """
    if self.check_logout(): return
    if organizationId is None: organizationId = self.organization
    if serviceTypeId is None: serviceTypeId = 'custom'
    result = self.ana_api.createService(organizationId=organizationId, name=name, description=description, volumes=volumes, instance=instance, serviceTypeId=serviceTypeId, tags=tags, editorId=serverId)
    return result


def edit_service(self, serviceId, name=None, description=None, volumes=None, instance=None, tags=None, schema=None):
    """Edit a service for your organization.
    
    Parameters
    ----------
    serviceId : str
        Service ID of the service to edit.
    name : name
        The new name to give the service.
    description : str
        Description of the service.
    volumes : list[str]
        Data volumes for the service.
    instance: str
        AWS Instance type to run the service on.
    tags : list[str]
        Tags for the service.
    schema : str
        Schema for the service.
    
    Returns
    -------
    bool
        If true, the service was successfully edited.
    """
    self.check_logout()
    if serviceId is None: raise Exception('serviceId must be specified.')
    result = self.ana_api.editService(serviceId=serviceId, name=name, description=description, volumes=volumes, instance=instance, tags=tags, schema=schema)
    return result


def delete_service(self, serviceId):
    """Delete a service from your organization.
    
    Parameters
    ----------
    serviceId : str
        Id of service to delete.
    
    Returns
    -------
    str
        Status
    """
    self.check_logout()
    if serviceId is None: raise Exception('ServiceId must be specified.')
    result = self.ana_api.deleteService(serviceId=serviceId)
    return result


def build_service(self, servicefile):
    """Build the Docker image of a service.
    
    Parameters
    ----------
    servicefile : str
        The service file for the service to build.

    Returns
    -------
    bool
        A boolean that indicates if the service Docker image was successfully built.
    """
    import os
    import docker
    import json
    import time
    import shutil
    import subprocess
    import tempfile
    from anatools.lib.print import print_color

    # make sure we can connect to docker
    try: dockerclient = docker.APIClient(base_url='unix://var/run/docker.sock')
    except: raise Exception('Cannot connect to Docker host.')    

    # check System and Docker space usage
    try:
        disk_usage = shutil.disk_usage('/')
        gb_free = disk_usage.free / (1024**3)  # Convert bytes to GB
        if self.verbose == 'debug': print(f"Disk space left: {gb_free:.3f}GB")
        if gb_free < 20: print_color(f'\nWarning: Low disk space detected! Only {gb_free:.1f}GB available.', 'ffff00')
        docker_space_cmd = "docker system df --format json"
        result = subprocess.run(docker_space_cmd, shell=True, capture_output=True, text=True)
        if result.returncode == 0:
            docker_stats = json.loads(result.stdout)
            total_docker_size = sum(item.get('Size', 0) for item in docker_stats if 'Size' in item)
            docker_gb = total_docker_size / (1024**3)  # Convert bytes to GB
            if self.verbose == 'debug': print(f"Docker space used: {docker_gb:.3f}GB")
            if docker_gb > 20:
                print_color(f'\nWarning: Docker is using {docker_gb:.1f}GB of disk space!', 'ffff00')
                print_color('Consider running "docker system prune -a" to free up space.', 'ffff00')
                print_color('This will remove:', 'ffff00')
                print_color('  - All stopped containers', 'ffff00')
                print_color('  - All networks not used by at least one container', 'ffff00')
                print_color('  - All dangling images', 'ffff00')
                print_color('  - All dangling build cache', 'ffff00')
    except Exception as e:
        pass

    # check for service Dockerfile
    print('Building Service Image...', end='', flush=True)    
    if not os.path.isfile(servicefile): raise Exception(f'No service file {servicefile} found.')
    servicedir, servicefile = os.path.split(servicefile)
    imgtime = int(time.time())
    image = f"anadeploy-service-{imgtime}"
    if servicedir == "": servicedir = "./"   
    if not os.path.isdir(os.path.join(servicedir, '.devcontainer')): raise Exception(f'No .devcontainer directory found in service directory.')
    if not os.path.isfile(os.path.join(servicedir, '.devcontainer/Dockerfile')): raise Exception(f'No Dockerfile found in .devcontainer/ directory, build requires Dockerfile.')

    # call the docker build command
    try:
        start = time.time()
        streamer = dockerclient.build(
            path=servicedir, 
            dockerfile=os.path.join(servicedir, '.devcontainer/Dockerfile'), 
            tag=image, 
            rm=True, 
            decode=True, 
            platform='linux/amd64', 
            nocache=False,
            pull=False)
        logfilepath = os.path.join(tempfile.gettempdir(), 'dockerbuild.log')
        logfile = open(logfilepath, 'w')
        while True:
            try:
                output = streamer.__next__()
                if self.verbose == 'debug':
                    if 'stream' in output: print(output['stream'].strip('\n'), flush=True)
                    if 'error' in output: print_color(f'{output["error"]}', 'ff0000')
                if 'stream' in output: logfile.write(output['stream'])
                if 'error' in output: logfile.write(output["error"])
                print(f'\rBuilding Service Image...  [{time.time()-start:.3f}s]', end='', flush=True)
            except StopIteration:
                time.sleep(1)
                try:
                    dockerclient = docker.from_env()
                    dockerclient.images.get(image)
                    logfile.close()
                    break
                except Exception as e: raise Exception(e)
    except Exception as e:
        logfile.close()
        raise Exception(f'Error encountered while building Docker image. Please check logfile {logfilepath}.')
    if self.verbose != 'debug' and os.path.exists(logfilepath): os.remove(logfilepath)
    print(f"\rBuilding Service Image...done.  [{time.time()-start:.3f}s]", flush=True)
    return image


def deploy_service(self, serviceId=None, servicefile=None):
    """Deploy the Docker image of a service.
    
    Parameters
    ----------
    serviceId : str
        ServiceId that you are pushing the image to.
    servicefile: str
        Name of the service file to look for. 
    
    Returns
    -------
    str
        deploymentId for current service deployment
    """
    import base64
    import docker
    import os
    import time

    self.check_logout()
    if serviceId is None and image is None: raise Exception('The serviceId or local image must be specified.')
    try: dockerclient = docker.from_env()
    except: raise Exception('Error connecting to Docker.')
    image = self.build_service(servicefile)
    try: serviceimage = dockerclient.images.get(image)
    except docker.errors.ImageNotFound: raise Exception(f'Could not find Docker image with name "{image}".')
    except: raise Exception('Error connecting to Docker.')

    dockerinfo = self.ana_api.deployService(serviceId)
    deploymentId = dockerinfo['deploymentId']
    reponame = dockerinfo['ecrEndpoint']
    encodedpass = dockerinfo['ecrPassword']
    if encodedpass:
        encodedbytes = encodedpass.encode('ascii')
        decodedbytes = base64.b64decode(encodedbytes)
        decodedpass = decodedbytes.decode('ascii').split(':')[-1]
    else: raise Exception('Failed to retrieve credentials from Rendered.ai platform.')

    # tag and push image
    serviceimage.tag(reponame)
    largest = 0
    start = time.time()
    servicedir, servicefile = os.path.split(servicefile)
    if servicedir == "": servicedir = "./" 
    logfilepath = os.path.join(servicedir, 'dockerpush.log')
    logfile = open(logfilepath, 'w')
    for line in dockerclient.images.push(reponame, auth_config={'username':'AWS', 'password':decodedpass}, stream=True, decode=True):
        logfile.write(str(line) + '\n')
        if 'status' in line and 'progressDetail' in line:
            if 'current' in line['progressDetail'] and 'total' in line['progressDetail']:
                progressDetail = line['progressDetail']
                if progressDetail['total'] >= largest:
                    largest = progressDetail['total']
                    print(f"\rPushing Service Image...  [{time.time()-start:.3f}s, {min(100,round((progressDetail['current']/progressDetail['total']) * 100))}%]", end='', flush=True)
        if 'error' in line: 
            if 'HTTP 403' in line['error']: raise Exception('You do not have permission to push to this repository.')
            else: raise Exception(line['error'])
    logfile.close()
    if self.verbose != 'debug' and os.path.exists(logfilepath): os.remove(logfilepath)
    print(f"\rPushing Service Image...done.  [{time.time()-start:.3f}s]     ", flush=True)
    
    # cleanup docker and update services
    dockerclient.images.remove(reponame)
    dockerclient.images.remove(image)
    dockerclient.close()
    return deploymentId


def get_service_deployment(self, deploymentId, stream=False):
    """Retrieves status for a service deployment.
    
    Parameters
    ----------
    deploymentId: str
        The deploymentId to retrieve status for
    stream: bool
        Flag to monitor deployment status until complete.

    Returns
    -------
    list[dict]
        Deployment status. 
    """
    import time
    self.check_logout()
    if deploymentId is None: raise Exception('DeploymentId must be specified.')
    if stream:
        data = self.ana_api.getServiceDeployment(deploymentId=deploymentId)
        print(f"\r\tStep {data['status']['step']} - {data['status']['message']}", end='', flush=True)
        while (data['status']['state'] not in ['Service Deployment Complete','Service Deployment Failed']):
            time.sleep(10)
            print(f"\r\tStep {data['status']['step']} - {data['status']['message']}", end='', flush=True)
            if self.check_logout(): return
            data = self.ana_api.getServiceDeployment(deploymentId=deploymentId)
        print(f"\r\tStep {data['status']['step']} - {data['status']['message']}", flush=True)
        return data
    else: return self.ana_api.getServiceDeployment(deploymentId=deploymentId)

    

def get_service_jobs(self, workspaceId=None, jobId=None, cursor=None, limit=None, filters=None, fields=None):
    """Fetches all service jobs that have run in a workspace.
    
    Parameters
    ----------
    workspaceId : str    
        The workspace to fetch the service jobs from. If not specified, the current workspace is used.
    jobId: str
        Filter service jobs list on the specific jobId.
    cursor : str
        Cursor for pagination.
    limit : int
        Maximum of service jobs to return.
    filters: dict
        Filters that limit output to entries that match the filter
    fields : list
        List of fields to return, leave empty to get all fields.
    
    Returns
    -------
    list[dict]
        List of service jobs associated with workspace.
    """
    self.check_logout()
    if workspaceId is None: workspaceId = self.workspace
    if limit is not None and limit <= 100: items = limit
    else: items = 100
    jobs = []
    while True:
        if limit and len(jobs) + items > limit: items = limit - len(jobs)
        ret = self.ana_api.getServiceJobs(workspaceId=workspaceId, jobId=jobId, limit=items, cursor=cursor, filters=filters, fields=fields)
        jobs.extend(ret)
        if len(ret) < items or len(jobs) == limit: break
        cursor = ret[-1]["serviceJobId"]
    return jobs


def create_service_job(self, serviceId, name, description=None, payload=None, workspaceId=None):
    """Create a service job for a service.
    
    Parameters
    ----------
    workspaceId : str
        Workspace ID of the workspace to create the job for. If not specified, the current workspace is used.
    serviceId : str
        Service ID of the service to create the job for.
    name : str
        Name of the job.
    description : str
        Description of the job.
    payload : dict
        Payload for the job.
    
    Returns
    -------
    str
        JobId of the created job.
    """
    self.check_logout()
    if workspaceId is None: workspaceId = self.workspace
    return self.ana_api.createServiceJob(workspaceId=workspaceId, serviceId=serviceId, name=name, description=description, payload=payload)


def delete_service_job(self, jobId, workspaceId=None):
    """Delete a service job for a service.
    
    Parameters
    ----------
    workspaceId : str
        Workspace ID of the workspace to delete the job from. If not specified, the current workspace is used.
    jobId : str
        JobId of the job to delete.
    
    Returns
    -------
    bool
        If true, the job was successfully deleted.
    """
    self.check_logout()
    if workspaceId is None: workspaceId = self.workspace
    return self.ana_api.deleteServiceJob(workspaceId=workspaceId, jobId=jobId)


def add_workspace_services(self, serviceIds, workspaceId=None):
    """Add a service to a workspace.
    
    Parameters
    ----------
    workspaceId : str
        Workspace ID of the workspace to add the service to. If not specified, the current workspace is used.
    serviceIds : str
        Service IDs of the services to add to the workspace.
    
    Returns
    -------
    bool
        If true, the services were successfully added to the workspace."""
    self.check_logout()
    if workspaceId is None: workspaceId = self.workspace
    return self.ana_api.addWorkspaceServices(workspaceId=workspaceId, serviceIds=serviceIds)


def remove_workspace_services(self, serviceIds, workspaceId=None):
    """Remove a service from a workspace.
    
    Parameters
    ----------
    workspaceId : str
        Workspace ID of the workspace to remove the service from. If not specified, the current workspace is used.
    serviceIds : str
        Service IDs of the services to remove from the workspace.
    
    Returns
    -------
    bool
        If true, the services were successfully removed from the workspace."""
    self.check_logout()
    if workspaceId is None: workspaceId = self.workspace
    return self.ana_api.removeWorkspaceServices(workspaceId=workspaceId, serviceIds=serviceIds)


def get_workspace_service_credentials(self, workspaceId=None, serviceId=None):
    """Get the credentials for a workspace service.

    Parameters
    ----------
    workspaceId : str
        Workspace ID of the workspace to get the service credentials for. If not specified, the current workspace is used.
    serviceId : str, optional
        Service ID of the service to get the credentials for.
        If not provided, the credentials for all services in the workspace are returned.
    
    Returns
    -------
    dict
        Dictionary of credentials for the workspace service."""
    self.check_logout()
    if workspaceId is None: workspaceId = self.workspace
    return self.ana_api.getWorkspaceServiceCredentials(workspaceId=workspaceId, serviceId=serviceId)
    
    
def get_instance_types(self):
    """Get the available instance types for a service.
    
    Returns
    -------
    list
        List of instance types."""
    self.check_logout()
    return self.ana_api.getInstanceTypes()


import os
from .datasets_test import get_dataset_files_fixture

### Tests ###
def test_get_image_annotation(client, get_dataset_files_fixture):
    files = get_dataset_files_fixture
    images = [file for file in files if 'images' in file['path']]
    annotation = client.get_image_annotation(datasetId=files[0]['datasetId'], filename=images[0], workspaceId=client.workspace)
    assert isinstance(annotation, dict)
    assert 'filename' in annotation

def test_get_image_mask(client, get_dataset_files_fixture):
    files = get_dataset_files_fixture
    images = [file for file in files if 'images' in file['path']]
    mask = client.get_image_mask(datasetId=files[0]['datasetId'], filename=images[0], workspaceId=client.workspace)
    assert isinstance(mask, str)
    assert os.path.isfile(mask)
    os.remove(mask)

def test_get_image_metadata(client, get_dataset_files_fixture):
    files = get_dataset_files_fixture
    images = [file for file in files if 'images' in file['path']]
    metadata = client.get_image_metadata(datasetId=files[0]['datasetId'], filename=images[0], workspaceId=client.workspace)
    assert isinstance(metadata, dict)
    assert 'filename' in metadata



import pytest


def test_get_api_keys(client):
    keys = client.get_api_keys()
    assert isinstance(keys, list)


# # TODO: support api key tests
# def test_get_api_key_data(client, test_create_api_key):
#     name, scope = test_create_api_key
#     data = client.get_api_key_data(name=name)
#     assert isinstance(data, dict)
#     assert data['name'] == name
#     assert data['scope'] == scope


# @pytest.fixture(scope="session")
# def test_create_api_key(client):
#     name = "anatools-api-key"
#     scope = "user"
#     key = client.create_api_key(name="anatools-api-key", scope="user")
#     assert isinstance(key, str) 
#     return name, scope


# @pytest.mark.dependency(depends=["test_get_api_key_data"])
# def test_delete_api_key(client, test_create_api_key):
#     name, scope = test_create_api_key
#     status = client.delete_api_key(name=name)
#     assert status == True



"""
Analytics Functions
"""

def get_analytics_types(self):
    """Retrieve the analytics types available on the Rendered.ai Platform.
    
    Returns
    -------
    list[str]
        The analytics types available on the Platform.
    """
    self.check_logout()
    return self.ana_api.getAnalyticsTypes()


def get_analytics(self, analyticsId=None, datasetId=None, workspaceId=None, cursor=None, limit=None, filters=None, fields=None):
    """Retrieve information about analytics jobs.
    
    Parameters
    ----------
    analyticsId : str
        Analytics Job ID.
    datasetId : str
        Dataset ID of the analytics job.
    workspaceId: str
        Workspace ID where the analytics exist. If none is provided, the current workspace will get used.
    cursor : str
        Cursor for pagination.
    limit : int
        Maximum number of analytics to return.
    filters: dict
        Filters that limit output to entries that match the filter 
    fields : list[str]
        List of fields to return, leave empty to get all fields.
    
    Returns
    -------
    list[dict]
        Analytics job information.
    """
    self.check_logout()
    if workspaceId is None: workspaceId = self.workspace
    if limit is not None and limit <= 100: items = limit
    else: items = 100
    analytics = []
    while True:
        if limit and len(analytics) + items > limit: items = limit - len(analytics)
        ret = self.ana_api.getAnalytics(workspaceId=workspaceId, datasetId=datasetId, analyticsId=analyticsId, limit=items, cursor=cursor, filters=filters, fields=fields)
        analytics.extend(ret)
        if len(ret) < items or len(analytics) == limit: break
        cursor = ret[-1]["analyticsId"]
    return analytics


def download_analytics(self, analyticsId, workspaceId=None):
    """Retrieve information about a specific analytics job. 
    If an analytics job is of type `Object Metrics` or `Mean Brightness`, then images will get downloaded to current working directory. 
    
    Parameters
    ----------
    analyticsId : str
        Analytics Job ID.
    workspaceId: str
        Workspace ID where the analytics exist. If none is provided, the current workspace will get used. 
    
    Returns
    -------
    list[dict]
        Analytics job information.
    """
    import os, json, requests
    self.check_logout()
    if analyticsId is None: raise ValueError("The analyticsId parameter is required!")
    if workspaceId is None: workspaceId = self.workspace
    result = self.ana_api.getAnalytics(workspaceId=workspaceId, analyticsId=analyticsId)
    if len(result) == 0: raise ValueError("Could not find analytics job with analyticsId: {analyticsId}")
    result = result[0]
    if result['type'] not in ['Object Metrics', 'Mean Brightness']: raise ValueError("Analytics job must be of type 'Object Metrics' or 'Mean Brightness' to download images.")
    analytics_result = json.dumps(result)
    presigned_urls = [str for str in analytics_result.split("\\\"") if str.startswith("https")]
    for i in range(len(presigned_urls)):
        filename = presigned_urls[i].split("/")[-1].split("?")[0]
        if self.interactive: print(f"\r[{i+1} / {len(presigned_urls)}]  Downloading {filename}...", end="", flush=True)
        with requests.get(presigned_urls[i], allow_redirects=True) as response:
            with open(filename, "wb") as outfile: outfile.write(response.content)
        analytics_result = analytics_result.replace(presigned_urls[i], os.path.join(os.getcwd(), filename), 1)
    if self.interactive: print("\r", end="")
    return json.loads(analytics_result)


def create_analytics(self, datasetId, type, workspaceId=None, tags=None):
    """Generate analytics for a dataset.
    
    Parameters
    ----------
    datasetId : str
        The datasetId of the dataset to generate analytics for.
    type : str
        The type of analytics to generate for the dataset.
    workspaceId : str
        Workspace ID of the dataset to generate the analytics for. If none is provided, the current workspace will get used. 
    tags : list[str]
        Tags for the analytics job.
    
    Returns
    -------
    str
        The analyticsId for the analytics job.
    """
    self.check_logout()
    if workspaceId is None: workspaceId = self.workspace
    if type not in self.get_analytics_types(): raise ValueError(f"The type parameter '{type}' is invalid, must be one of {self.get_analytics_types()}.")
    return self.ana_api.createAnalytics(workspaceId=workspaceId, datasetId=datasetId, type=type, tags=tags)


def delete_analytics(self, analyticsId, workspaceId=None):
    """Deletes analytics for a dataset.
    
    Parameters
    ----------
    analyticsId : str
        Analytics ID for the analytics to delete. 
    workspaceId: str
        Workspace ID where the analytics exist. If none is provided, the current workspace will get used. 
    
    Returns
    -------
    bool
        If true, successfully deleted the analytics.
    """
    self.check_logout()
    if analyticsId is None: raise ValueError("The analyticsId parameter is required!")
    if workspaceId is None: workspaceId = self.workspace
    return self.ana_api.deleteAnalytics(workspaceId=workspaceId, analyticsId=analyticsId)


def edit_analytics(self, analyticsId, tags=None, workspaceId=None):
    """Edits analytics for a dataset.
    
    Parameters
    ----------
    analyticsId : str
        Analytics ID for the analytics to edit. 
    tags : list[str]
        Tags for the analytics job.
    workspaceId: str
        Workspace ID where the analytics exist. If none is provided, the current workspace will get used. 
    
    Returns
    -------
    bool
        If true, successfully edited the analytics.
    """
    self.check_logout()
    if analyticsId is None: raise ValueError("The analyticsId parameter is required!")
    if not isinstance(tags, list): raise ValueError("The tags parameter must be a list!")
    if workspaceId is None: workspaceId = self.workspace
    return self.ana_api.editAnalytics(workspaceId=workspaceId, analyticsId=analyticsId, tags=tags)
    


"""
Channels Functions
"""

def get_channels(self, organizationId=None, workspaceId=None, channelId=None, cursor=None, limit=None, filters=None, fields=None):
    """Fetches all 
    
    Parameters
    ----------
    organizationId : str
        Filter channel list on what's available to the organization.
    workspaceId : str    
        Filter channel list on what's available to the workspace.
    channelId: str
        Filter channel list on the specific channelId.
    cursor : str
        Cursor for pagination.
    limit : int
        Maximum of channels to return.
    filters: dict
        Filters that limit output to entries that match the filter
    fields : list
        List of fields to return, leave empty to get all fields.
    
    Returns
    -------
    list[dict]
        List of channels associated with user, workspace, organization or channelId.
    """
    self.check_logout()
    if limit is not None and limit <= 100: items = limit
    else: items = 100
    channels = []
    while True:
        if limit and len(channels) + items > limit: items = limit - len(channels)
        ret = self.ana_api.getChannels(organizationId=organizationId, workspaceId=workspaceId, channelId=channelId, limit=items, cursor=cursor, filters=filters, fields=fields)
        channels.extend(ret)
        if len(ret) < items or len(channels) == limit: break
        cursor = ret[-1]["channelId"]
    return channels
                

def get_channel_nodes(self, channelId, fields=None):
    """Get the nodes for a channel.
    
    Parameters
    ----------
    channelId : str
        Channel Id to filter.
    fields : list
        List of fields to return, leave empty to get all fields.
   
    Returns
    -------
    list[dict]
        channel node schema data
    """
    import json
    self.check_logout()
    response = self.ana_api.getChannelSchema(channelId=channelId)
    data = json.loads(response)
    if fields: data = data = [{k: v for k, v in item.items() if k in fields} for item in data]
    return data


def create_channel(self, name, description=None, organizationId=None, volumes=[], instance=None, timeout=120, interfaceVersion=1):
    """Create a channel for your organization.
    
    Parameters
    ----------
    name : str
        Channel name.
    description : str
        Description of the channel
    organizationId : str
        Organization ID. Defaults to current if not specified.
    volumes : list[str]
        List of the data volume names to associate with this channel.
    instance: str
        AWS Instance type.
    timeout: int
        Maximum runtime of a channel run.
    interface: int
        The ana interface version number.
   
    Returns
    -------
    list[dict]
        channel data
    """
    if self.check_logout(): return
    if organizationId is None: organizationId = self.organization
    result = self.ana_api.createChannel(organizationId=organizationId, name=name, description=description, volumes=volumes, instance=instance, timeout=timeout, interfaceVersion=interfaceVersion)
    return result


def edit_channel(self, channelId, name=None, description=None, volumes=None, instance=None, timeout=None, status=None, interfaceVersion=None, preview=None):
    """Edit a channel for your organization.
    
    Parameters
    ----------
    channelId : str
        ChannelId ID of the channel to edit.
    name : name
        The new name to give the channel.
    description : str
        Description of the channel
    volumes : list[str]
        Data volumes for the channel.
    instance: str
        Instance type to run the channel on.
    timeout: int
        Maximum runtime for the channel run.
    status: str
        The status of the channel.
    interface: int
        The ana interface version number.
    preview: bool
        Enable or disable the preview for the channel.
    
    Returns
    -------
    bool
        If true, the channel was successfully edited.
    """
    self.check_logout()
    if channelId is None: raise Exception('ChannelId must be specified.')
    result = self.ana_api.editChannel(channelId=channelId, name=name, description=description, volumes=volumes, instance=instance, timeout=timeout, status=status, interfaceVersion=interfaceVersion, preview=preview)
    return result


def delete_channel(self, channelId):
    """Delete a channel of your organization.
    
    Parameters
    ----------
    channelId : str
        Id of channel to delete.
    
    Returns
    -------
    str
        Status
    """
    self.check_logout()
    if channelId is None: raise Exception('ChannelId must be specified.')
    result = self.ana_api.deleteChannel(channelId=channelId)
    return result


def build_channel(self, channelfile, ignore=['data/', 'output/'], verify=False):
    """Build the Docker image of a channel.
    
    Parameters
    ----------
    channelfile : str
        The channel file for the channel to build.
    ignore : list, optional
        List of files or directories to ignore during the build.
    verify : bool, optional
        If True, verifies the image by running the anautils command.

    Returns
    -------
    bool
        A boolean that indicates if the channel Docker image was successfully built.
    """
    import os
    import docker
    import json
    import shutil
    import time
    import subprocess
    from anatools.lib.print import print_color

    # make sure we can connect to docker
    try: dockerclient = docker.APIClient(base_url='unix://var/run/docker.sock')
    except: raise Exception('Cannot connect to Docker host.')

    # check System and Docker space usage
    try:
        disk_usage = shutil.disk_usage('/')
        gb_free = disk_usage.free / (1024**3)  # Convert bytes to GB
        if self.verbose == 'debug': print(f"Disk space left: {gb_free:.3f}GB")
        if gb_free < 20: print_color(f'\nWarning: Low disk space detected! Only {gb_free:.1f}GB available.', 'ffff00')
        docker_space_cmd = "docker system df --format json"
        result = subprocess.run(docker_space_cmd, shell=True, capture_output=True, text=True)
        if result.returncode == 0:
            docker_stats = json.loads(result.stdout)
            total_docker_size = sum(item.get('Size', 0) for item in docker_stats if 'Size' in item)
            docker_gb = total_docker_size / (1024**3)  # Convert bytes to GB
            if self.verbose == 'debug': print(f"Docker space used: {docker_gb:.3f}GB")
            if docker_gb > 20:
                print_color(f'\nWarning: Docker is using {docker_gb:.1f}GB of disk space!', 'ffff00')
                print_color('Consider running "docker system prune -a" to free up space.', 'ffff00')
                print_color('This will remove:', 'ffff00')
                print_color('  - All stopped containers', 'ffff00')
                print_color('  - All networks not used by at least one container', 'ffff00')
                print_color('  - All dangling images', 'ffff00')
                print_color('  - All dangling build cache', 'ffff00')
    except Exception as e:
        print(e)
        pass

    # build dockerfile
    print('Building Channel Image...', end='', flush=True)    
    if not os.path.isfile(channelfile): raise Exception(f'No channel file {channelfile} found.')
    channeldir, channelfile = os.path.split(channelfile)
    if channeldir == "": channeldir = "./"   
    if not os.path.isdir(os.path.join(channeldir, '.devcontainer')): raise Exception(f'No .devcontainer directory found in channel directory.')
    if not os.path.isfile(os.path.join(channeldir, '.devcontainer/Dockerfile')): raise Exception(f'Issue detected with .devcontainer directory, build requires Dockerfile.')

    # check if dockerfile already exists, if so rename it
    if os.path.isfile(os.path.join(channeldir, 'Dockerfile')):
        os.rename(os.path.join(channeldir, 'Dockerfile'), os.path.join(channeldir, 'Dockerfile.old'))

    # create new Dockerfile
    with open(os.path.join(channeldir, 'Dockerfile'), 'w+') as buildfile:
        with open(os.path.join(channeldir,'.devcontainer/Dockerfile')) as dockerfile: buildfile.write(dockerfile.read())
        deploycommands = """\n
# Commands used for deployment
WORKDIR /ana
COPY . .
RUN sudo chown -R $(whoami):$(whoami) /ana
RUN sudo /home/$USERNAME/miniconda3/bin/pip install -r /ana/requirements.txt || \\
    sudo /home/$USERNAME/.conda/bin/pip install -r /ana/requirements.txt
RUN sudo /home/$USERNAME/miniconda3/envs/anatools/bin/pip install -r /ana/requirements.txt || \\
    sudo /home/$USERNAME/.conda/envs/anatools/bin/pip install -r /ana/requirements.txt 
"""
        buildfile.write(deploycommands)

    # build the dockerignore for docker context, ignore specified files and directories
    if os.path.isfile(os.path.join(channeldir, '.dockerignore')):
        os.rename(os.path.join(channeldir, '.dockerignore'), os.path.join(channeldir, '.dockerignore.old'))
    with open(os.path.join(channeldir, '.dockerignore'), 'w+') as ignorefile:
        for i in ignore: ignorefile.write(f'{i}\n')

    # call the docker build command
    status = False
    try:
        start = time.time()
        streamer = dockerclient.build(path=channeldir, tag=channelfile.split('.')[0], decode=True)
        logfile = open(os.path.join(channeldir, 'dockerbuild.log'), 'w')
        while True:
            try:
                output = streamer.__next__()
                if self.verbose == 'debug':
                    if 'stream' in output: print(output['stream'].strip('\n'), flush=True)
                    if 'error' in output: print_color(f'{output["error"]}', 'ff0000')
                if 'stream' in output: logfile.write(output['stream'])
                if 'error' in output: logfile.write(output["error"])
                print(f'\rBuilding Channel Image...  [{time.time()-start:.3f}s]', end='', flush=True)
            except StopIteration as e:
                time.sleep(5)
                print(f"\rBuilding Channel Image...done.  [{time.time()-start:.3f}s]", flush=True)
                try:
                    dockerclient = docker.from_env()
                    dockerclient.images.get(channelfile.split('.')[0])
                    status = True
                    logfile.close()
                    break
                except Exception as e:
                    raise Exception()
    except Exception as e:
        raise Exception(f'Error encountered while building Docker image. Please check logfile dockerbuild.log.')

    # if verify, check that image is valid by running anautils command
    if verify:
        try:
            print(f'Verifying Channel Image...', end='', flush=True)
            start = time.time()

            # Run anautils command in the Docker container to verify schema
            container_name = f"verify_{channelfile.split('.')[0]}_{int(time.time())}"
            cmd = [
                "docker", "run", "--name", container_name, 
                channelfile.split('.')[0], 
                "anautils", "--mode=schema", "--output=/tmp"
            ]
            
            result = subprocess.run(cmd,  capture_output=True, text=True, check=False)
            subprocess.run(["docker", "rm", container_name], capture_output=True)
            if result.returncode != 0:
                print_color(f"\n\n{result.stderr}\n", "ff0000")
                raise Exception(f'Error encountered while verifying Docker image with the anautils command.')
            
            print(f"\rVerifying Channel Image...done. [{time.time()-start:.3f}s]")
            status = True
                
        except Exception as e:
            raise Exception(f'Error encountered while verifying Docker image. Please check that you can generate the schema using the anautils command.')

    # cleanup
    if self.verbose != 'debug': os.remove(os.path.join(channeldir, 'dockerbuild.log'))
    os.remove(os.path.join(channeldir, '.dockerignore'))
    os.remove(os.path.join(channeldir, 'Dockerfile'))
    if os.path.isfile(os.path.join(channeldir, '.dockerignore.old')):
        os.rename(os.path.join(channeldir, '.dockerignore.old'), os.path.join(channeldir, '.dockerignore'))
    if os.path.isfile(os.path.join(channeldir, 'Dockerfile.old')):
        os.rename(os.path.join(channeldir, 'Dockerfile.old'), os.path.join(channeldir, 'Dockerfile'))
    
    return status


def deploy_channel(self, channelId=None, channelfile=None, image=None):
    """Deploy the Docker image of a channel.
    
    Parameters
    ----------
    channelId : str
        ChannelId that you are pushing the image to. If the channelId isn't specified, it will use the image name to lookup the channelId.
    channelfile: str
        Name of the channel file to look for. 
    image: str
        The Docker image name. This should match the channel name when running ana. If image is not specified, it will use the channel name for the channelId.
    
    Returns
    -------
    str
        deploymentId for current round of deployment or an error message if something went wrong
    """
    import os
    import docker
    import time
    import base64

    self.check_logout()
    if channelId is None and image is None: raise Exception('The channelId or local image must be specified.')
    try: dockerclient = docker.from_env()
    except: raise Exception('Cannot connect to Docker host.')

    # build channel image if not specified
    if channelfile and not image:
        if self.build_channel(channelfile, verify=True): image = os.path.split(channelfile)[1].split('.')[0]
        else: return False
    else: raise Exception('The channelfile or a docker image must be specified!')
    try: channelimage = dockerclient.images.get(image)
    except docker.errors.ImageNotFound: raise Exception(f'Could not find Docker image with name "{image}".')
    except: raise Exception('Error connecting to Docker.')

    # get repository info
    print(f"Pushing Channel Image...", end='', flush=True)
    dockerinfo = self.ana_api.deployChannel(channelId, image)
    deploymentId = dockerinfo['deploymentId']
    reponame = dockerinfo['ecrEndpoint']
    encodedpass = dockerinfo['ecrPassword']
    if encodedpass:
        encodedbytes = encodedpass.encode('ascii')
        decodedbytes = base64.b64decode(encodedbytes)
        decodedpass = decodedbytes.decode('ascii').split(':')[-1]
    else: raise Exception('Failed to retrieve credentials from Rendered.ai platform.')

    # tag and push image
    channelimage.tag(reponame)
    largest = 0
    start = time.time()
    channeldir, channelfile = os.path.split(channelfile)
    if channeldir == "": channeldir = "./"  
    logfile = open(os.path.join(channeldir, 'dockerpush.log'), 'w')
    for line in dockerclient.images.push(reponame, auth_config={'username':'AWS', 'password':decodedpass}, stream=True, decode=True):
        if 'status' in line and 'progressDetail' in line:
            if 'current' in line['progressDetail'] and 'total' in line['progressDetail']:
                progressDetail = line['progressDetail']
                if progressDetail['total'] >= largest:
                    largest = progressDetail['total']
                    print(f"\rPushing Channel Image...  [{time.time()-start:.3f}s, {min(100,round((progressDetail['current']/progressDetail['total']) * 100))}%]", end='', flush=True)
        logfile.write(str(line) + '\n')
    logfile.close()
    if self.verbose != 'debug': os.remove(os.path.join(channeldir, 'dockerpush.log'))
    print(f"\rPushing Channel Image...done.  [{time.time()-start:.3f}s]     ", flush=True)
    
    # cleanup docker and update channels
    dockerclient.images.remove(reponame)
    dockerclient.images.remove(image)
    dockerclient.close()
    return deploymentId


def get_deployment_status(self, deploymentId, stream=False):
    """Retrieves status for a channel's deployment.
    
    Parameters
    ----------
    deploymentId: str
        The deploymentId to retrieve status for
    stream: bool
        Flag to print information to the terminal so the user can avoid constant polling to retrieve status.

    Returns
    -------
    list[dict]
        Deployment status. 
    """
    import time
    self.check_logout()
    if deploymentId is None: raise Exception('DeploymentId must be specified.')
    if stream:
        data = self.ana_api.getChannelDeployment(deploymentId=deploymentId)
        print(f"\r\tStep {data['status']['step']} - {data['status']['message']}", end='', flush=True)
        while (data['status']['state'] not in ['Channel Deployment Complete','Channel Deployment Failed']):
            time.sleep(10)
            print(f"\r\tStep {data['status']['step']} - {data['status']['message']}", end='', flush=True)
            if self.check_logout(): return
            data = self.ana_api.getChannelDeployment(deploymentId=deploymentId)
        print(f"\r\tStep {data['status']['step']} - {data['status']['message']}", flush=True)
        return data
    else: return self.ana_api.getChannelDeployment(deploymentId=deploymentId)

    
def get_channel_documentation(self, channelId):
    """Returns channel documentation as markdown text.
    
    Parameters
    ----------
    channelID: str
        The channelId of the channel

    Returns
    -------
    str
        The markdown file for channel documentation.
    """
    return self.ana_api.getChannelDocumentation(channelId=channelId)
    

def upload_channel_documentation(self, channelId, mdfile):
    """Uploads a markdown file for channel documentation.
    
    Parameters
    ----------
    channelID: str
        The channelId of the channel
    mdfile: str
        The filepath of the markdown file used for channel documentation.

    Returns
    -------
    bool
        Success/Failure of channel documenation upload.
    """
    import os
    import requests

    if not os.path.isfile(mdfile): raise ValueError(f'Could not find file {mdfile}')
    if os.path.splitext(mdfile)[1] != '.md': raise ValueError('The channel documentation file must be in markdown format with .md extension.') 
    fileinfo = self.ana_api.uploadChannelDocumentation(channelId=channelId, keys=[os.path.basename(mdfile)])[0]
    with open(mdfile, 'rb') as filebytes:
        files = {'file': filebytes}
        data = {
            "key":                  fileinfo['fields']['key'],
            "bucket":               fileinfo['fields']['bucket'],
            "X-Amz-Algorithm":      fileinfo['fields']['algorithm'],
            "X-Amz-Credential":     fileinfo['fields']['credential'],
            "X-Amz-Date":           fileinfo['fields']['date'],
            "X-Amz-Security-Token": fileinfo['fields']['token'],
            "Policy":               fileinfo['fields']['policy'],
            "X-Amz-Signature":      fileinfo['fields']['signature'],
        }
        response = requests.post(fileinfo['url'], data=data, files=files)
        if response.status_code != 204: 
            print(response.status_code)
            raise Exception('Failed to upload channel documentation file.')
    return True


def get_node_documentation(self, channelId, node, fields=None):
    """Retrieves the markdown documentation for a node.
    
    Parameters
    ----------
    channelId: str
        The channelId of the channel
    node: str
        The node to retrieve documentation for.
    fields: list[str]
        List of fields to retrieve for the node documentation.

    Returns
    -------
    str
        The markdown documentation for the node.
    """
    data = self.ana_api.getNodeDocumentation(channelId=channelId, node=node, fields=fields)
    if data: return data
    else: return False
    

def profile_channel(self, graphId=None, channelId=None, instances=None, workspaceId=None):
    """Profile a channel using a graph on the Rendered.ai Platform.
    
    Parameters
    ----------
    name: str
        Name for dataset. 
    graphId : str
        ID of the staged graph to create dataset from.
    description : str 
        Description for new dataset.
    runs : int
        Number of times a channel will run within a single job. This is also how many different images will get created within the dataset. 
    priority : int
        Job priority.
    seed : int
        Seed number.
    workspaceId : str
        Workspace ID of the staged graph's workspace. If none is provided, the current workspace will get used. 
    
    Returns
    -------
    str
        Success or failure message about dataset creation.
    """
    import datetime
    import json
    import time

    if self.check_logout(): return
    if channelId is None and graphId is None: raise ValueError('Either the channelId or graphId parameters must be specified!')
    if instances is None: instances = []
    if type(instances) != list: raise ValueError('The instances parameter must be a list of valid AWS instances types!')
    if workspaceId is None: workspaceId = self.workspace
    
    timestamp = datetime.datetime.now().isoformat()
    starttime = time.time()
    created_graph = False
    if graphId:
        try:
            graphs = self.ana_api.getGraphs(graphId=graphId)
            if len(graphs) == 0: raise ValueError(f'Could not find graph with ID "{graphId}"')
            graphId = graphs[0]['graphId']
            channel = self.ana_api.getChannels(channelId=graphs[0]['channelId'])
        except Exception as e: raise e
    else:
        try:
            channels = self.ana_api.getChannels(channelId=channelId)
            if len(channels) == 0: raise ValueError(f'Could not find channel with ID "{channelId}"')
            channel = channels[0]
            graph = self.ana_api.getDefaultGraph(channelId=channelId)
            graphId = self.ana_api.createGraph(name=f'default-{timestamp}', channelId=channel['channelId'], description=f"Staged graph for profiling channel {channel['name']} (channelId={channel['channelId']}).", staged=True, graph=graph, workspaceId=workspaceId)
            created_graph = True
        except Exception as e: raise e

    # create dataset jobs
    if self.interactive: print(f"Profiling channel {channel['name']}...", end='\r', flush=True)
    if not instances: instances = ['g4dn.xlarge', 'g5.xlarge', 'g6.xlarge', 'g6e.xlarge'] # self.ana_api.getDefaultInstanceTypes()
    jobs = {}
    for instance in instances:
        datasetId = self.ana_api.createDataset(workspaceId=workspaceId, graphId=graphId, name=f"profile-{timestamp}", description=f"Profile of channel {channel['name']} (channelId={channel['channelId']}) on instance type {instance}", runs=1, seed=0, priority=1, instanceType=instance)
        jobs[instance] = {'datasetId': datasetId}

    # wait for dataset jobs to complete
    complete = []
    while True:
        for instance in jobs.keys():
            if jobs[instance]['datasetId'] in complete: continue
            jobdata = self.ana_api.getDatasets(datasetId=jobs[instance]['datasetId'], workspaceId=workspaceId)[0]
            if jobdata['status'] in ['success', 'failed', 'timeout']:
                complete.append(jobs[instance]['datasetId'])
                jobs[instance]['status'] = jobdata['status']
        time.sleep(10)
        if self.interactive: print(f"Profiling channel {channel['name']}...\t[{len(complete)}/{len(jobs)} {int(time.time()-starttime)}]s", end='\r', flush=True)
        if len(jobs.keys()) == len(complete): break
    
    # get dataset metrics
    for instance in jobs.keys():
        runs = self.ana_api.getDatasetRuns(datasetId=jobs[instance]['datasetId'], workspaceId=workspaceId)
        jobs[instance]['metrics'] = self.ana_api.getDatasetMetrics(datasetId=jobs[instance]['datasetId'], runId=runs[0]['runId'], workspaceId=workspaceId)
        # jobs[instance]['stats'] = self.ana_api.getInstanceStats(instance=instance)
    if self.interactive: print(f"Profiling channel {channel['name']}...done! [{int(time.time()-starttime)}s]\033[K", flush=True)

    return jobs


"""
Organization Functions
"""


def set_organization(self, organizationId, workspaceId=None):
    """Set the organization (and optionally a workspace) to the one you wish to work in.

    Parameters
    ----------
    organizationId : str
        Organization ID for the organization you wish to work in.
    workspaceId : str, optional
        Workspace ID for the workspace you wish to work in. Uses default workspace if this is not set.
    """
    from anatools.lib.print import print_color

    self.check_logout()
    if organizationId is None: raise Exception('OrganizationId must be specified.')
    organization = self.get_organizations(organizationId=organizationId)[0]
    self.organization = organization['organizationId']
    workspaces = self.get_workspaces(organizationId=self.organization)
    if len(workspaces) == 0: 
        if self.interactive: raise Exception("No workspaces available, please contact support@rendered.ai."); 
    if workspaceId:
        if workspaceId not in [w['workspaceId'] for w in workspaces]:
            raise Exception(f"Workspace {workspaceId} not found in organization {organizationId}.")
        self.workspace = workspaceId
    self.workspace = workspaces[0]['workspaceId']
    if self.interactive: print_color(f'The current organization is: {self.organization}\nThe current workspace is: {self.workspace}', '91e600')
    return True


def get_organizations(self, organizationId=None, cursor=None, limit=None, fields=None):
    """Shows the organizations the user belongs to and the user's role in that organization.
    
    Parameters
    ----------
    organizationId : str
        Organization ID to filter.
    cursor : str
        Cursor for pagination.
    limit : int
        Maximum number of organizations to return.
    fields : list
        List of fields to return, leave empty to get all fields.
    
    Returns
    -------
    list[dict]
        Information about the organizations you belong to. 
    """  
    self.check_logout()
    if limit is not None and limit <= 100: items = limit
    else: items = 100
    organizations = []
    while True:
        if limit and len(organizations) + items > limit: items = limit - len(organizations)
        ret = self.ana_api.getOrganizations(organizationId, limit=items, cursor=cursor, fields=fields)
        organizations.extend(ret)
        if len(ret) < items or len(organizations) == limit: break
        cursor = ret[-1]["organizationId"]
    return organizations


def edit_organization(self, name, organizationId=None):
    """Update the organization name. Uses current organization if no organizationId provided.
    
    Parameters
    ----------
    name : str
        Name to update organization to.
    organizationId : str
        Organization Id to update.
    
    Returns
    -------
    bool
        True if organization was edited successfully, False otherwise.
    """  
    self.check_logout()
    if name is None: raise ValueError("The name parameter is required!")
    if organizationId is None: organizationId = self.organization
    return self.ana_api.editOrganization(organizationId=organizationId, name=name)


def get_organization_members(self, organizationId=None, cursor=None, limit=None, fields=None):
    """Get users of an organization.
    
    Parameters
    ----------
    organizationId : str
        Organization ID. Defaults to current if not specified.
    cursor : str
        Cursor for pagination.
    limit : int
        Maximum number of members to return.
    fields : list
        List of fields to return, leave empty to get all fields.
    
    Returns
    -------
    list[dict]
        Information about users of an organization.
    """
    self.check_logout()
    if organizationId is None: organizationId = self.organization
    if limit is not None and limit <= 100: items = limit
    else: items = 100
    members = []
    while True:
        if limit and len(members) + items > limit: items = limit - len(members)
        ret = self.ana_api.getMembers(organizationId=organizationId, limit=items, cursor=cursor, fields=fields)
        members.extend(ret)
        if len(ret) < items or len(members) == limit: break
        cursor = ret[-1]["userId"]
    return members


def get_organization_invites(self, organizationId=None, cursor=None, limit=None, fields=None):
    """Get invitations of an organization.
    
    Parameters
    ----------
    organizationId : str
        Organization ID. Defaults to current if not specified.
    cursor : str
        Cursor for pagination.
    limit : int
        Maximum number of invitations to return.
    fields : list
        List of fields to return, leave empty to get all fields.
    
    Returns
    -------
    list[dict]
        Information about invitations of an organization.
    """
    self.check_logout()
    if organizationId is None: organizationId = self.organization
    if limit is not None and limit <= 100: items = limit
    else: items = 100
    invites = []
    while True:
        if limit and len(invites) + items > limit: items = limit - len(invites)
        ret = self.ana_api.getInvitations(organizationId=organizationId, limit=items, cursor=cursor, fields=fields)
        invites.extend(ret)
        if len(ret) < items or len(invites) == limit: break
        cursor = ret[-1]["email"]
    return invites


def add_organization_member(self, email, role, organizationId=None):
    """Add a user to an existing organization.
    
    Parameters
    ----------
    email: str
        Email of user to add.
    role : str
        Role for user. 
    organizationId : str
        Organization ID to add members too. Uses current if not specified.
    
    Returns
    -------
    str
        Response status if user got added to workspace succesfully. 
    """
    self.check_logout()
    if email is None: raise ValueError("The email parameter is required!")
    if role is None: raise ValueError("The role parameter is required!")
    if organizationId is None: organizationId = self.organization
    return self.ana_api.addMember(email=email, role=role, organizationId=organizationId)


def remove_organization_member(self, email, organizationId=None):
    """Remove a member from an existing organization.
    
    Parameters
    ----------
    email : str
        Member email to remove.
    organizationId: str
        Organization ID to remove member from. Removes from current organization if not specified.
    
    Returns
    -------
    str
        Response status if member got removed from organization succesfully. 
    """
    self.check_logout()
    if email is None: raise ValueError("The email parameter is required!")
    if organizationId is None: organizationId = self.organization
    return self.ana_api.removeMember(email=email, organizationId=organizationId)


def remove_organization_invitation(self, email, organizationId=None):
    """Remove a invitation from an existing organization.
    
    Parameters
    ----------
    email : str
        Invitation email to remove.
    organizationId: str
        Organization ID to remove member from. Removes from current organization if not specified.
    
    Returns
    -------
    str
        Response status if member got removed from organization succesfully. 
    """
    self.check_logout()
    if email is None: raise ValueError("The email parameter is required!")
    if organizationId is None: organizationId = self.organization
    return self.ana_api.removeMember(email=email, organizationId=organizationId)


def edit_organization_member(self, email, role, organizationId=None):
    """Edit a member's role. 
    
    Parameters
    ----------
    email : str
        Member email to edit.
    role: str
        Role to assign. 
    organizationId: str
        Organization ID to remove member from. Edits member in current organization if not specified.
    
    Returns
    -------
    str
        Response if member got edited succesfully. 
    """
    self.check_logout()
    if email is None: raise ValueError("The email parameter is required!")
    if role is None: raise ValueError("The role parameter is required!")
    if organizationId is None: organizationId = self.organization
    return self.ana_api.editMember(email=email, role=role, organizationId=organizationId)



"""
Large Language Model Functions
"""

def get_llm_response(self, promptId, fields=None):
    """Retrieves the response to an LLM prompt.
    
    Parameters
    ----------
    promptId : str
        The ID of a prompt.
    fields : list[str], optional
        The fields to retrieve from the response.
    
    Returns
    -------
    dict
        Prompt response info
    """
    if self.check_logout(): return
    llm_response = self.ana_api.getLLMResponse(promptId=promptId, fields=fields)
    return llm_response


def create_llm_prompt(self, prompt, baseChannel, nodeType, nodeName):
    """ Creates an LLM prompt.
    
    Parameters
    ----------
    prompt : str
        The prompt to create

    baseChannel: str
        The base channel to use for examples

    nodeType: str
        The type of node to create

    nodeName: str
        The name of the node to create

    Returns
    -------
    str
        Prompt ID
    """
    if self.check_logout(): return
    prompt_id = self.ana_api.createLLMPrompt(prompt=prompt, baseChannel=baseChannel, nodeType=nodeType, nodeName=nodeName)
    return prompt_id


def delete_llm_prompt(self, promptId):
    """Deletes an LLM prompt.
    
    Parameters
    ----------
    promptId : str
        The ID of a prompt.
    
    Returns
    -------
    bool
        Success code
    """
    if self.check_logout(): return
    llm_response = self.ana_api.deleteLLMPrompt(promptId=promptId)
    return llm_response


def get_llm_base_channels(self):
    """Gets a list of the base channels

    Returns
    -------
    list[str]
        A list of the base channels
    """
    if self.check_logout(): return
    llm_response = self.ana_api.getLLMBaseChannels()
    return llm_response


def get_llm_channel_node_types(self):
    """Gets a dictionary of base channels. For each channel there is a list of valid node types.

    Returns
    -------
    dict
        A dictionary of base channels and their valid node types
    """
    if self.check_logout(): return
    llm_response = self.ana_api.getLLMChannelNodeTypes()
    return llm_response